# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Tuple, Union

import torch
import transformers
from megatron.core.transformer.transformer_config import TransformerConfig
from transformers import AutoConfig as HFAutoConfig
from transformers import AutoModelForImageTextToText
from transformers import Qwen2_5_VLConfig as HFQwen25VLConfig
from transformers import Qwen2VLConfig as HFQwen2VLConfig
from transformers import Qwen2VLForConditionalGeneration
from transformers.models.qwen2_5_vl.configuration_qwen2_5_vl import Qwen2_5_VLVisionConfig as HFQwen25VLVisionConfig
from transformers.models.qwen2_vl.configuration_qwen2_vl import Qwen2VLVisionConfig as HFQwen2VLVisionConfig

from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec
from nemo.collections.llm import (
    Qwen2Config,
    Qwen2Config1P5B,
    Qwen2Config7B,
    Qwen2Config72B,
    Qwen25Config3B,
    Qwen25Config7B,
    Qwen25Config32B,
    Qwen25Config72B,
)
from nemo.collections.vlm.neva.model.llava import export_qkv, export_qkv_bias
from nemo.collections.vlm.qwen2vl.model.base import (
    Qwen2VLConfig,
    Qwen2VLModel,
    Qwen2VLVisionConfig,
    Qwen25VLVisionConfig,
)
from nemo.collections.vlm.vision import MultimodalProjectorConfig
from nemo.export.trt_llm.nemo_ckpt_loader.nemo_file import load_distributed_model_weights
from nemo.lightning import io, teardown
from nemo.lightning.io.state import _ModelState
from nemo.lightning.pytorch.utils import dtype_from_hf
from nemo.utils import logging

if TYPE_CHECKING:
    from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer


# Note: these Qwen2VL configs are copied from the corresponding HF model. You may need to modify the parameter for
# your own needs
@dataclass
class Qwen2VLConfig2B(Qwen2VLConfig):
    """Qwen2VL Config 2B"""

    from transformers import PretrainedConfig

    language_transformer_config: TransformerConfig = field(
        default_factory=lambda: Qwen2Config1P5B(share_embeddings_and_output_weights=True)
    )
    vision_transformer_config: Union[TransformerConfig, PretrainedConfig] = field(
        default_factory=lambda: Qwen2VLVisionConfig(num_layers=32, num_attention_heads=16)
    )
    vision_projection_config: TransformerConfig = field(
        default_factory=lambda: MultimodalProjectorConfig(input_size=5120, hidden_size=1536, ffn_hidden_size=5120)
    )


@dataclass
class Qwen2VLConfig7B(Qwen2VLConfig):
    """Qwen2VL Config 7B"""

    from transformers import PretrainedConfig

    language_transformer_config: TransformerConfig = field(default_factory=lambda: Qwen2Config7B())
    vision_transformer_config: Union[TransformerConfig, PretrainedConfig] = field(
        default_factory=lambda: Qwen2VLVisionConfig(num_layers=32, num_attention_heads=16)
    )
    vision_projection_config: TransformerConfig = field(
        default_factory=lambda: MultimodalProjectorConfig(input_size=5120, hidden_size=3584, ffn_hidden_size=5120)
    )


@dataclass
class Qwen2VLConfig72B(Qwen2VLConfig):
    """Qwen2VL Config 72B"""

    from transformers import PretrainedConfig

    language_transformer_config: TransformerConfig = field(default_factory=lambda: Qwen2Config72B())
    vision_transformer_config: Union[TransformerConfig, PretrainedConfig] = field(
        default_factory=lambda: Qwen2VLVisionConfig(num_layers=32, num_attention_heads=16)
    )
    vision_projection_config: TransformerConfig = field(
        default_factory=lambda: MultimodalProjectorConfig(input_size=5120, hidden_size=8192, ffn_hidden_size=5120)
    )


@dataclass
class Qwen25VLConfig3B(Qwen2VLConfig):
    """Qwen2.5VL Config 3B"""

    from transformers import PretrainedConfig

    language_transformer_config: TransformerConfig = field(default_factory=lambda: Qwen25Config3B())
    vision_transformer_config: Union[TransformerConfig, PretrainedConfig] = field(
        default_factory=lambda: Qwen25VLVisionConfig(num_layers=32, num_attention_heads=16)
    )
    vision_projection_config: TransformerConfig = field(
        default_factory=lambda: MultimodalProjectorConfig(
            projector_type="mcore_mlp", input_size=5120, hidden_size=2048, ffn_hidden_size=5120
        )
    )


@dataclass
class Qwen25VLConfig7B(Qwen2VLConfig):
    """Qwen2.5VL Config 7B"""

    from transformers import PretrainedConfig

    language_transformer_config: TransformerConfig = field(default_factory=lambda: Qwen25Config7B())
    vision_transformer_config: Union[TransformerConfig, PretrainedConfig] = field(
        default_factory=lambda: Qwen25VLVisionConfig(num_layers=32, num_attention_heads=16)
    )
    vision_projection_config: TransformerConfig = field(
        default_factory=lambda: MultimodalProjectorConfig(
            projector_type="mcore_mlp", input_size=5120, hidden_size=3584, ffn_hidden_size=5120
        )
    )


@dataclass
class Qwen25VLConfig32B(Qwen2VLConfig):
    """Qwen2.5VL Config 32B"""

    from transformers import PretrainedConfig

    language_transformer_config: TransformerConfig = field(default_factory=lambda: Qwen25Config32B())
    vision_transformer_config: Union[TransformerConfig, PretrainedConfig] = field(
        default_factory=lambda: Qwen25VLVisionConfig(num_layers=32, num_attention_heads=16, ffn_hidden_size=3456)
    )
    vision_projection_config: TransformerConfig = field(
        default_factory=lambda: MultimodalProjectorConfig(
            projector_type="mcore_mlp", input_size=5120, hidden_size=5120, ffn_hidden_size=5120
        )
    )


@dataclass
class Qwen25VLConfig72B(Qwen2VLConfig):
    """Qwen2.5VL Config 72B"""

    from transformers import PretrainedConfig

    language_transformer_config: TransformerConfig = field(default_factory=lambda: Qwen25Config72B())
    vision_transformer_config: Union[TransformerConfig, PretrainedConfig] = field(
        default_factory=lambda: Qwen25VLVisionConfig(num_layers=32, num_attention_heads=16, ffn_hidden_size=3456)
    )
    vision_projection_config: TransformerConfig = field(
        default_factory=lambda: MultimodalProjectorConfig(
            projector_type="mcore_mlp", input_size=5120, hidden_size=8192, ffn_hidden_size=5120
        )
    )


@io.model_importer(Qwen2VLModel, "hf")
class HFQwen2VLImporter(io.ModelConnector["Qwen2VLForConditionalGeneration", Qwen2VLModel]):
    """Qwen2VL Model HF Importer"""

    def init(self) -> Qwen2VLModel:
        # pylint: disable=C0115,C0116
        return Qwen2VLModel(self.config, model_version="qwen2-vl", tokenizer=self.tokenizer)

    def apply(self, output_path: Path) -> Path:
        # pylint: disable=C0115,C0116
        source = AutoModelForImageTextToText.from_pretrained(str(self), trust_remote_code=True)
        hf_config = HFAutoConfig.from_pretrained(str(self), trust_remote_code=True)
        self.is_v2_5 = hf_config.model_type == "qwen2_5_vl"

        target = self.init()
        trainer = self.nemo_setup(target)
        source = source.to(dtype_from_hf(hf_config))
        target = target.to(dtype_from_hf(hf_config))
        self.convert_state(source, target)
        print(f"Converted Qwen2VL model to Nemo, saving to {output_path}")
        self.nemo_save(output_path, trainer)
        print(f"Converted Qwen2VL model saved to {output_path}")

        teardown(trainer, target)
        del trainer, target

        return output_path

    def convert_state(self, source, target):
        # pylint: disable=C0115,C0116,C0301
        mapping = {
            "visual.patch_embed.proj.weight": "vision_model.conv1.weight",
            "visual.blocks.*.norm1.weight": "vision_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight",
            "visual.blocks.*.norm1.bias": "vision_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_bias",
            "visual.blocks.*.norm2.weight": "vision_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight",
            "visual.blocks.*.norm2.bias": "vision_model.decoder.layers.*.mlp.linear_fc1.layer_norm_bias",
            "visual.blocks.*.attn.proj.weight": "vision_model.decoder.layers.*.self_attention.linear_proj.weight",
            "visual.blocks.*.attn.proj.bias": "vision_model.decoder.layers.*.self_attention.linear_proj.bias",
            "model.embed_tokens.weight": "language_model.embedding.word_embeddings.weight",
            "model.layers.*.self_attn.o_proj.weight": "language_model.decoder.layers.*.self_attention.linear_proj.weight",
            "model.layers.*.mlp.down_proj.weight": "language_model.decoder.layers.*.mlp.linear_fc2.weight",
            "model.layers.*.input_layernorm.weight": "language_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight",
            "model.layers.*.post_attention_layernorm.weight": "language_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight",
            "model.norm.weight": "language_model.decoder.final_layernorm.weight",
            # "lm_head.weight": "language_model.output_layer.weight",
        }
        if not target.config.language_transformer_config.share_embeddings_and_output_weights:
            mapping.update({"lm_head.weight": "language_model.output_layer.weight"})

        if self.is_v2_5:
            mapping.update(
                {
                    "visual.blocks.*.mlp.down_proj.weight": "vision_model.decoder.layers.*.mlp.linear_fc2.weight",
                    "visual.blocks.*.mlp.down_proj.bias": "vision_model.decoder.layers.*.mlp.linear_fc2.bias",
                    "visual.merger.ln_q.weight": "vision_model.decoder.final_layernorm.weight",
                }
            )
        else:
            mapping.update(
                {
                    "visual.blocks.*.mlp.fc1.weight": "vision_model.decoder.layers.*.mlp.linear_fc1.weight",
                    "visual.blocks.*.mlp.fc1.bias": "vision_model.decoder.layers.*.mlp.linear_fc1.bias",
                    "visual.blocks.*.mlp.fc2.weight": "vision_model.decoder.layers.*.mlp.linear_fc2.weight",
                    "visual.blocks.*.mlp.fc2.bias": "vision_model.decoder.layers.*.mlp.linear_fc2.bias",
                    "visual.merger.ln_q.weight": "vision_model.decoder.final_layernorm.weight",
                    "visual.merger.ln_q.bias": "vision_model.decoder.final_layernorm.bias",
                }
            )
        if "vision_projection.encoder.linear_fc1.weight" in target.module.state_dict().keys():
            mapping.update(
                {
                    "visual.merger.mlp.0.weight": "vision_projection.encoder.linear_fc1.weight",
                    "visual.merger.mlp.0.bias": "vision_projection.encoder.linear_fc1.bias",
                    "visual.merger.mlp.2.weight": "vision_projection.encoder.linear_fc2.weight",
                    "visual.merger.mlp.2.bias": "vision_projection.encoder.linear_fc2.bias",
                }
            )
        elif "vision_projection.0.weight" in target.module.state_dict().keys():
            mapping.update(
                {
                    "visual.merger.mlp.0.weight": "vision_projection.0.weight",
                    "visual.merger.mlp.0.bias": "vision_projection.0.bias",
                    "visual.merger.mlp.2.weight": "vision_projection.2.weight",
                    "visual.merger.mlp.2.bias": "vision_projection.2.bias",
                }
            )
        else:
            raise KeyError("Unable to map vision projection keys.")

        transforms = [
            _import_language_qkv,
            _import_language_qkv_bias,
            _import_vision_qkv,
            _import_vision_qkv_bias,
            _import_linear_fc1,
        ]
        if self.is_v2_5:
            transforms += [
                _import_vision_linear_fc1_weight,
                _import_vision_linear_fc1_bias,
            ]
        return io.apply_transforms(
            source,
            target,
            mapping=mapping,
            transforms=transforms,
        )

    @property
    def tokenizer(self) -> "AutoTokenizer":
        # pylint: disable=C0115,C0116
        from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer

        return AutoTokenizer(str(self))

    @property
    def config(self) -> Qwen2VLConfig:
        # pylint: disable=C0115,C0116
        from packaging.version import Version

        if Version(transformers.__version__) > Version('4.51.3'):
            # Todo: need to fix with newest version of transformers
            raise ValueError(
                f"Current version of transformers is {transformers.__version__},"
                f"Please lower the version to be <= 4.51.3"
            )

        hf_config = HFAutoConfig.from_pretrained(str(self), trust_remote_code=True)
        is_v2_5 = hf_config.model_type == "qwen2_5_vl"

        def make_vocab_size_divisible_by(vocab_size):
            # pylint: disable=C0115,C0116
            base = 128
            while vocab_size % base != 0:
                base //= 2
            return base

        text_config = hf_config
        language_transformer_config = Qwen2Config(
            num_layers=text_config.num_hidden_layers,
            hidden_size=text_config.hidden_size,
            ffn_hidden_size=text_config.intermediate_size,
            num_attention_heads=text_config.num_attention_heads,
            init_method_std=text_config.initializer_range,
            layernorm_epsilon=text_config.rms_norm_eps,
            num_query_groups=text_config.num_key_value_heads,
            rotary_base=text_config.rope_theta,
            gated_linear_unit=True,
            make_vocab_size_divisible_by=make_vocab_size_divisible_by(text_config.vocab_size),
            share_embeddings_and_output_weights=text_config.tie_word_embeddings,
            vocab_size=text_config.vocab_size,
            fp16=(dtype_from_hf(text_config) == torch.float16),
            bf16=(dtype_from_hf(text_config) == torch.bfloat16),
            params_dtype=dtype_from_hf(text_config),
        )

        # Use MCore instead of Pytorch
        vision_config = hf_config.vision_config
        if is_v2_5:
            vision_transformer_config = Qwen25VLVisionConfig(
                ffn_hidden_size=vision_config.intermediate_size,
                fp16=(dtype_from_hf(hf_config) == torch.float16),
                bf16=(dtype_from_hf(hf_config) == torch.bfloat16),
                params_dtype=dtype_from_hf(hf_config),
            )
            merge_hidden_size = vision_config.hidden_size * (vision_config.spatial_merge_size**2)
            vision_projection_config = MultimodalProjectorConfig(
                input_size=merge_hidden_size,
                hidden_size=vision_config.out_hidden_size,
                ffn_hidden_size=merge_hidden_size,
                projector_type="mcore_mlp",
                fp16=(dtype_from_hf(hf_config) == torch.float16),
                bf16=(dtype_from_hf(hf_config) == torch.bfloat16),
                params_dtype=dtype_from_hf(hf_config),
            )
        else:
            vision_transformer_config = Qwen2VLVisionConfig(
                fp16=(dtype_from_hf(hf_config) == torch.float16),
                bf16=(dtype_from_hf(hf_config) == torch.bfloat16),
                params_dtype=dtype_from_hf(hf_config),
            )
            merge_hidden_size = vision_config.embed_dim * (vision_config.spatial_merge_size**2)
            vision_projection_config = MultimodalProjectorConfig(
                input_size=merge_hidden_size,
                hidden_size=vision_config.hidden_size,
                ffn_hidden_size=merge_hidden_size,
                projector_type="mcore_mlp",
                fp16=(dtype_from_hf(hf_config) == torch.float16),
                bf16=(dtype_from_hf(hf_config) == torch.bfloat16),
                params_dtype=dtype_from_hf(hf_config),
            )

        output = Qwen2VLConfig(
            language_transformer_config=language_transformer_config,
            vision_transformer_config=vision_transformer_config,
            vision_projection_config=vision_projection_config,
            vision_feature_layer=-1,
            fp16=(dtype_from_hf(hf_config) == torch.float16),
            bf16=(dtype_from_hf(hf_config) == torch.bfloat16),
            params_dtype=dtype_from_hf(hf_config),
        )

        return output


@io.model_exporter(Qwen2VLModel, "hf")
class HFQwen2VLExporter(io.ModelConnector[Qwen2VLModel, "Qwen2VLForConditionalGeneration"]):
    """
    Exporter class for converting NeMo Qwen2VL model to HuggingFace format.

    Inherits:
        io.ModelConnector: Connector interface to handle setup, save, and load using the Lightning framework.

    Methods:
        init: Initializes a new HuggingFace Qwen2VL model instance.
        apply: Converts the NeMo model to HuggingFace format and saves it.
        convert_state: Maps and transforms the state dictionary from NeMo to HuggingFace format.
        config: Generates and returns the HuggingFace Qwen2VL config for the model.
    """

    def init(self, dtype=torch.bfloat16) -> "Qwen2VLForConditionalGeneration":
        """
        Initializes a HuggingFace Qwen2VLForConditionalGeneration model.

        Args:
            dtype: The data type to use for the model (default: torch.bfloat16)

        Returns:
            Qwen2VLForConditionalGeneration: A HuggingFace Qwen2VL model initialized with the configuration.
        """
        from transformers.modeling_utils import no_init_weights

        with no_init_weights():
            return AutoModelForImageTextToText.from_config(self.config, torch_dtype=dtype)

    def apply(self, output_path: Path) -> Path:
        """
        Converts the NeMo Qwen2VL model to HuggingFace format and saves it to the specified path.

        Args:
            output_path (Path): The path where the converted HuggingFace model will be saved.

        Returns:
            Path: The output path where the HuggingFace model was saved.
        """
        logging.info("Loading Qwen2VL NeMo checkpoint. This may take a while...")
        source, source_config = self.ckpt_load(self)
        logging.info("Qwen2VL NeMo checkpoint loaded.")
        logging.info("Initializing the HF model..")
        target = self.init()
        logging.info("Start Converting the model..")
        target = self.convert_state(source, target, source_config)
        target = target.cpu()
        target.save_pretrained(output_path)

        try:
            self.tokenizer.tokenizer.save_pretrained(output_path)
        except Exception:
            logging.warning("Failed to save tokenizer")

        print(f"Converted Qwen2VL model saved to {output_path}")

        return output_path

    def convert_state(self, source, target, source_config):
        # pylint: disable=C0115,C0116,line-too-long
        """
        Maps and transforms the state dictionary from NeMo to HuggingFace format.

        Args:
            source: The source NeMo model.
            target: The target HuggingFace model.

        Returns:
            The target HuggingFace model with the converted state.
        """

        mapping = {
            "vision_model.conv1.weight": "visual.patch_embed.proj.weight",
            "vision_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight": "visual.blocks.*.norm1.weight",
            "vision_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_bias": "visual.blocks.*.norm1.bias",
            "vision_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight": "visual.blocks.*.norm2.weight",
            "vision_model.decoder.layers.*.mlp.linear_fc1.layer_norm_bias": "visual.blocks.*.norm2.bias",
            "vision_model.decoder.layers.*.self_attention.linear_proj.weight": "visual.blocks.*.attn.proj.weight",
            "vision_model.decoder.layers.*.self_attention.linear_proj.bias": "visual.blocks.*.attn.proj.bias",
            "language_model.embedding.word_embeddings.weight": "model.embed_tokens.weight",
            "language_model.decoder.layers.*.self_attention.linear_proj.weight": "model.layers.*.self_attn.o_proj.weight",
            "language_model.decoder.layers.*.mlp.linear_fc2.weight": "model.layers.*.mlp.down_proj.weight",
            "language_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight": "model.layers.*.input_layernorm.weight",
            "language_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight": "model.layers.*.post_attention_layernorm.weight",
            "language_model.decoder.final_layernorm.weight": "model.norm.weight",
            # "language_model.output_layer.weight": "lm_head.weight",
        }
        if source_config.language_transformer_config.share_embeddings_and_output_weights:
            mapping.update({"language_model.embedding.word_embeddings.weight": "lm_head.weight"})
        else:
            mapping.update({"language_model.output_layer.weight": "lm_head.weight"})

        if self.is_v2_5:
            mapping.update(
                {
                    "vision_model.decoder.layers.*.mlp.linear_fc2.weight": "visual.blocks.*.mlp.down_proj.weight",
                    "vision_model.decoder.layers.*.mlp.linear_fc2.bias": "visual.blocks.*.mlp.down_proj.bias",
                    "vision_model.decoder.final_layernorm.weight": "visual.merger.ln_q.weight",
                }
            )

        else:
            mapping.update(
                {
                    "vision_model.decoder.layers.*.mlp.linear_fc1.weight": "visual.blocks.*.mlp.fc1.weight",
                    "vision_model.decoder.layers.*.mlp.linear_fc1.bias": "visual.blocks.*.mlp.fc1.bias",
                    "vision_model.decoder.layers.*.mlp.linear_fc2.weight": "visual.blocks.*.mlp.fc2.weight",
                    "vision_model.decoder.layers.*.mlp.linear_fc2.bias": "visual.blocks.*.mlp.fc2.bias",
                    "vision_model.decoder.final_layernorm.weight": "visual.merger.ln_q.weight",
                    "vision_model.decoder.final_layernorm.bias": "visual.merger.ln_q.bias",
                }
            )
        if "vision_projection.encoder.linear_fc1.weight" in source.state_dict().keys():
            mapping.update(
                {
                    "vision_projection.encoder.linear_fc1.weight": "visual.merger.mlp.0.weight",
                    "vision_projection.encoder.linear_fc1.bias": "visual.merger.mlp.0.bias",
                    "vision_projection.encoder.linear_fc2.weight": "visual.merger.mlp.2.weight",
                    "vision_projection.encoder.linear_fc2.bias": "visual.merger.mlp.2.bias",
                }
            )
        elif "vision_projection.0.weight" in source.state_dict().keys():
            mapping.update(
                {
                    "vision_projection.0.weight": "visual.merger.mlp.0.weight",
                    "vision_projection.0.bias": "visual.merger.mlp.0.bias",
                    "vision_projection.2.weight": "visual.merger.mlp.2.weight",
                    "vision_projection.2.bias": "visual.merger.mlp.2.bias",
                }
            )
        else:
            raise KeyError("Unable to map vision projection keys.")

        transforms = [
            _export_language_qkv,
            _export_language_qkv_bias,
            _export_vision_qkv,
            _export_vision_qkv_bias,
            _export_linear_fc1,
        ]
        if self.is_v2_5:
            transforms += [
                _export_vision_linear_fc1_weight,
                _export_vision_linear_fc1_bias,
            ]

        return io.apply_transforms(
            source,
            target,
            mapping=mapping,
            transforms=transforms,
        )

    @property
    def tokenizer(self) -> "TokenizerSpec":
        """
        Gets the tokenizer from the loaded model context.

        Returns:
            The tokenizer specification.
        """
        return io.load_context(str(self), subpath="model").tokenizer

    def ckpt_load(self, path: Path) -> Tuple[Dict, Dict]:
        """
        This function loads the state dict directly from a distributed checkpoint, and modify the state dict
        so that it is consistent with the key names you would get from loading the checkpoint into a model.
        This is a more memory-efficient method to obtain a state dict without initializing the nemo model.

        Args:
            path (Path): The path from which the model will be loaded.

        Returns
        -------
            Tuple[Dict, Dict]: The loaded state dict and the yaml config dict.
        """
        config = io.load_context(str(self), subpath="model.config")
        dist_ckpt_folder = path / "weights"
        state_dict = {}

        langauge_layers = config.language_transformer_config.num_layers
        vision_layers = config.vision_transformer_config.num_layers
        distributed_model_weights = load_distributed_model_weights(dist_ckpt_folder, True).items()
        for k, v in distributed_model_weights:
            if "_extra_state" in k:
                continue
            new_k = k.replace("module.", "")
            if "layers" in new_k and (v.size(0) == langauge_layers or v.size(0) == vision_layers):
                # Only split layers
                for i in range(v.size(0)):
                    state_dict[new_k.replace("layers", f"layers.{str(i)}")] = v[i]
            state_dict[new_k] = v

        source = _ModelState(state_dict)
        return source, config

    @property
    def config(self) -> "HFQwen2VLConfig":
        """
        Generates the configuration for the HuggingFace Qwen2VL model based on the NeMo model.

        Returns:
            HFQwen2VLConfig: A configuration object for the HuggingFace Qwen2VL model.
        """
        from packaging.version import Version

        if Version(transformers.__version__) > Version('4.51.3'):
            # Todo: need to fix with newest version of transformers
            raise ValueError(
                f"Current version of transformers is {transformers.__version__},"
                f"Please lower the version to be <= 4.51.3"
            )
        source = io.load_context(str(self), subpath="model.config")

        language_config = source.language_transformer_config
        vision_model_config = source.vision_transformer_config
        vision_projection_config = source.vision_projection_config

        self.is_v2_5 = hasattr(vision_model_config, "fullatt_block_indexes") and (
            vision_model_config.fullatt_block_indexes != None
        )

        if self.is_v2_5:
            vision_config = HFQwen25VLVisionConfig(
                depth=vision_model_config.num_layers,
                embed_dim=vision_model_config.embed_dim,
                hidden_size=vision_model_config.hidden_size,
                out_hidden_size=language_config.hidden_size,
                hidden_act="silu",
                mlp_ratio=int(vision_projection_config.ffn_hidden_size // vision_model_config.hidden_size),
                num_heads=vision_model_config.num_attention_heads,
                in_channels=3,
                patch_size=vision_model_config.patch_dim,
                spatial_merge_size=vision_model_config.spatial_merge_size,
                spatial_patch_size=vision_model_config.spatial_patch_size,
                temporal_patch_size=vision_model_config.temporal_patch_size,
                initializer_range=vision_model_config.init_method_std,
                fullatt_block_indexes=[7, 15, 23, 31],
                tokens_per_second=2,
                model_type="qwen2_5_vl",
                torch_dtype="bfloat16",
            ).to_dict()

            # Create the LlavaConfig for HuggingFace
            hf_config = HFQwen25VLConfig(
                vision_config=vision_config,
                num_hidden_layers=language_config.num_layers,
                hidden_size=language_config.hidden_size,
                intermediate_size=language_config.ffn_hidden_size,
                num_attention_heads=language_config.num_attention_heads,
                max_window_layers=70,
                max_position_embeddings=language_config.seq_length,
                initializer_range=language_config.init_method_std,
                rms_norm_eps=language_config.layernorm_epsilon,
                num_key_value_heads=language_config.num_query_groups,
                rope_theta=language_config.rotary_base,
                vocab_size=language_config.vocab_size,
                rope_scaling={"type": "mrope", "mrope_section": [16, 24, 24]},
                tie_word_embeddings=language_config.share_embeddings_and_output_weights,
                torch_dtype="bfloat16",
                # vocab_size=self.tokenizer.vocab_size,
                bos_token_id=151643,
                eos_token_id=151645,
                vision_start_token_id=151652,
                vision_end_token_id=151653,
                vision_token_id=151654,
                image_token_id=151655,
                video_token_id=51656,
            )
            return hf_config
        else:
            vision_config = HFQwen2VLVisionConfig(
                depth=vision_model_config.num_layers,
                embed_dim=vision_model_config.embed_dim,
                hidden_size=vision_projection_config.hidden_size,
                hidden_act="quick_gelu",
                mlp_ratio=int(vision_projection_config.ffn_hidden_size // vision_model_config.hidden_size),
                num_heads=vision_model_config.num_attention_heads,
                in_channels=3,
                patch_size=vision_model_config.patch_dim,
                spatial_merge_size=vision_model_config.spatial_merge_size,
                spatial_patch_size=vision_model_config.spatial_patch_size,
                temporal_patch_size=vision_model_config.temporal_patch_size,
                initializer_range=vision_model_config.init_method_std,
                model_type="qwen2_vl",
                torch_dtype="bfloat16",
            ).to_dict()

            # Create the Qwen2VLConfig for HuggingFace
            # if transformers > 4.51.3, use Qwen2VLTextConfig as text_config
            # https://github.com/huggingface/transformers/pull/37268
            return HFQwen2VLConfig(
                num_hidden_layers=language_config.num_layers,
                hidden_size=language_config.hidden_size,
                intermediate_size=language_config.ffn_hidden_size,
                num_attention_heads=language_config.num_attention_heads,
                initializer_range=language_config.init_method_std,
                rms_norm_eps=language_config.layernorm_epsilon,
                num_key_value_heads=language_config.num_query_groups,
                rope_theta=language_config.rotary_base,
                tie_word_embeddings=language_config.share_embeddings_and_output_weights,
                vocab_size=language_config.vocab_size,
                vision_config=vision_config,
                torch_dtype="bfloat16",
            )


def import_qkv(q, k, v, head_num, num_query_groups, heads_per_group, hidden_size, head_size):
    # pylint: disable=C0115,C0116
    old_tensor_shape = q.size()
    new_q_tensor_shape = (head_num, head_size) + old_tensor_shape[1:]
    new_kv_tensor_shape = (num_query_groups, head_size) + old_tensor_shape[1:]

    q = q.view(*new_q_tensor_shape)
    k = k.view(*new_kv_tensor_shape)
    v = v.view(*new_kv_tensor_shape)

    qkv_weights_l = []
    for i in range(num_query_groups):
        qkv_weights_l.append(q[i * heads_per_group : (i + 1) * heads_per_group, :, :])
        qkv_weights_l.append(k[i : i + 1, :, :])
        qkv_weights_l.append(v[i : i + 1, :, :])
    qkv_weights = torch.cat(qkv_weights_l)
    assert qkv_weights.ndim == 3, qkv_weights.shape
    assert qkv_weights.shape[0] == (heads_per_group + 2) * num_query_groups, qkv_weights.shape
    assert qkv_weights.shape[1] == head_size, qkv_weights.shape
    assert qkv_weights.shape[2] == old_tensor_shape[1], qkv_weights.shape

    qkv_weights = qkv_weights.reshape([head_size * (head_num + 2 * num_query_groups), hidden_size])

    return qkv_weights


@io.state_transform(
    source_key=("visual.blocks.*.attn.qkv.weight",),
    target_key="vision_model.decoder.layers.*.self_attention.linear_qkv.weight",
)
def _import_vision_qkv(ctx: io.TransformCTX, hf_qkv_weights):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config.vision_transformer_config

    slice = int(hf_qkv_weights.shape[0] / 3)
    assert slice == megatron_config.hidden_size
    q = hf_qkv_weights[:slice, :]
    k = hf_qkv_weights[slice : slice * 2, :]
    v = hf_qkv_weights[slice * 2 :, :]

    return import_qkv(
        q,
        k,
        v,
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=megatron_config.hidden_size,
        head_size=megatron_config.kv_channels,
    )


@io.state_transform(
    source_key=("visual.blocks.*.attn.qkv.bias",),
    target_key="vision_model.decoder.layers.*.self_attention.linear_qkv.bias",
)
def _import_vision_qkv_bias(ctx: io.TransformCTX, hf_qkv_bias):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config.vision_transformer_config

    slice = int(hf_qkv_bias.shape[0] / 3)
    assert slice == megatron_config.hidden_size

    q_bias = hf_qkv_bias[:slice]
    k_bias = hf_qkv_bias[slice : slice * 2]
    v_bias = hf_qkv_bias[slice * 2 :]

    return import_qkv(
        q_bias.unsqueeze(-1),
        k_bias.unsqueeze(-1),
        v_bias.unsqueeze(-1),
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=1,
        head_size=megatron_config.kv_channels,
    ).squeeze(-1)


@io.state_transform(
    source_key=(
        "model.layers.*.self_attn.q_proj.weight",
        "model.layers.*.self_attn.k_proj.weight",
        "model.layers.*.self_attn.v_proj.weight",
    ),
    target_key="language_model.decoder.layers.*.self_attention.linear_qkv.weight",
)
def _import_language_qkv(ctx: io.TransformCTX, q, k, v):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config.language_transformer_config
    return import_qkv(
        q,
        k,
        v,
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=megatron_config.hidden_size,
        head_size=megatron_config.kv_channels,
    )


@io.state_transform(
    source_key=(
        "model.layers.*.self_attn.q_proj.bias",
        "model.layers.*.self_attn.k_proj.bias",
        "model.layers.*.self_attn.v_proj.bias",
    ),
    target_key="language_model.decoder.layers.*.self_attention.linear_qkv.bias",
)
def _import_language_qkv_bias(ctx: io.TransformCTX, q_bias, k_bias, v_bias):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config.language_transformer_config
    return import_qkv(
        q_bias.unsqueeze(-1),
        k_bias.unsqueeze(-1),
        v_bias.unsqueeze(-1),
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=1,
        head_size=megatron_config.kv_channels,
    ).squeeze(-1)


@io.state_transform(
    source_key=("vision_model.embeddings.class_embedding",),
    target_key="vision_model.class_token",
)
def _import_cls_token(ctx: io.TransformCTX, cls_token):
    # pylint: disable=C0115,C0116
    return cls_token.reshape(1, 1, -1)


@io.state_transform(
    source_key=(
        "model.layers.*.mlp.gate_proj.weight",
        "model.layers.*.mlp.up_proj.weight",
    ),
    target_key="language_model.decoder.layers.*.mlp.linear_fc1.weight",
)
def _import_linear_fc1(down, gate):
    # pylint: disable=C0115,C0116
    return torch.cat((down, gate), axis=0)


@io.state_transform(
    source_key=("visual.blocks.*.mlp.gate_proj.weight", "visual.blocks.*.mlp.up_proj.weight"),
    target_key="vision_model.decoder.layers.*.mlp.linear_fc1.weight",
)
def _import_vision_linear_fc1_weight(down, gate):
    # pylint: disable=C0115,C0116
    return torch.cat((down, gate), axis=0)


@io.state_transform(
    source_key=("visual.blocks.*.mlp.gate_proj.bias", "visual.blocks.*.mlp.up_proj.bias"),
    target_key="vision_model.decoder.layers.*.mlp.linear_fc1.bias",
)
def _import_vision_linear_fc1_bias(down, gate):
    # pylint: disable=C0115,C0116
    return torch.cat((down, gate), axis=0)


def export_qkv(linear_qkv, head_num, num_query_groups, heads_per_group, hidden_size, head_size):
    # pylint: disable=C0115,C0116
    qkv_total_dim = head_num + 2 * num_query_groups

    linear_qkv = linear_qkv.reshape([qkv_total_dim, head_size, -1])
    hidden_size = linear_qkv.size(-1)
    q_slice = torch.cat(
        [
            torch.arange((heads_per_group + 2) * i, (heads_per_group + 2) * i + heads_per_group)
            for i in range(num_query_groups)
        ]
    )
    k_slice = torch.arange(heads_per_group, qkv_total_dim, (heads_per_group + 2))
    v_slice = torch.arange(heads_per_group + 1, qkv_total_dim, (heads_per_group + 2))

    q_proj = linear_qkv[q_slice].reshape(-1, hidden_size).cpu()
    k_proj = linear_qkv[k_slice].reshape(-1, hidden_size).cpu()
    v_proj = linear_qkv[v_slice].reshape(-1, hidden_size).cpu()

    return q_proj, k_proj, v_proj


def export_qkv_bias(qkv_bias: torch.Tensor, head_num, num_query_groups, heads_per_group, head_size):
    """
    Split interleave-concatenated qkv bias to separate q, k, v bias

    Example: export layer linear_qkv bias to HF {q|k|v}_proj bias
    """
    qkv_total_dim = head_num + 2 * num_query_groups

    qkv_bias = qkv_bias.reshape([qkv_total_dim, head_size])
    q_slice = torch.cat(
        [
            torch.arange((heads_per_group + 2) * i, (heads_per_group + 2) * i + heads_per_group)
            for i in range(num_query_groups)
        ]
    )
    k_slice = torch.arange(heads_per_group, qkv_total_dim, (heads_per_group + 2))
    v_slice = torch.arange(heads_per_group + 1, qkv_total_dim, (heads_per_group + 2))

    q_bias = qkv_bias[q_slice].reshape(-1).cpu()
    k_bias = qkv_bias[k_slice].reshape(-1).cpu()
    v_bias = qkv_bias[v_slice].reshape(-1).cpu()

    return q_bias, k_bias, v_bias


@io.state_transform(
    source_key="vision_model.decoder.layers.*.self_attention.linear_qkv.weight",
    target_key="visual.blocks.*.attn.qkv.weight",
)
def _export_vision_qkv(ctx: io.TransformCTX, qkv):
    # pylint: disable=C0115,C0116
    hf_config = ctx.target.config.vision_config
    hidden_size = hf_config.embed_dim if hf_config.model_type == "qwen2_vl" else hf_config.hidden_size
    return torch.cat(
        export_qkv(
            qkv,
            head_num=hf_config.num_heads,
            num_query_groups=hf_config.num_heads,
            heads_per_group=hf_config.num_heads // hf_config.num_heads,
            hidden_size=hidden_size,
            head_size=hidden_size // hf_config.num_heads,
        ),
        axis=0,
    )


@io.state_transform(
    source_key="vision_model.decoder.layers.*.self_attention.linear_qkv.bias",
    target_key="visual.blocks.*.attn.qkv.bias",
)
def _export_vision_qkv_bias(ctx: io.TransformCTX, qkv_bias):
    # pylint: disable=C0115,C0116
    hf_config = ctx.target.config.vision_config
    hidden_size = hf_config.embed_dim if hf_config.model_type == "qwen2_vl" else hf_config.hidden_size
    return torch.cat(
        export_qkv_bias(
            qkv_bias,
            head_num=hf_config.num_heads,
            num_query_groups=hf_config.num_heads,
            heads_per_group=hf_config.num_heads // hf_config.num_heads,
            head_size=hidden_size // hf_config.num_heads,
        ),
        axis=0,
    )


@io.state_transform(
    source_key="language_model.decoder.layers.*.self_attention.linear_qkv.weight",
    target_key=(
        "model.layers.*.self_attn.q_proj.weight",
        "model.layers.*.self_attn.k_proj.weight",
        "model.layers.*.self_attn.v_proj.weight",
    ),
)
def _export_language_qkv(ctx: io.TransformCTX, qkv):
    # pylint: disable=C0115,C0116
    hf_config = ctx.target.config
    return export_qkv(
        qkv,
        head_num=hf_config.num_attention_heads,
        num_query_groups=hf_config.num_key_value_heads,
        heads_per_group=hf_config.num_attention_heads // hf_config.num_key_value_heads,
        hidden_size=hf_config.hidden_size,
        head_size=hf_config.hidden_size // hf_config.num_attention_heads,
    )


@io.state_transform(
    source_key="language_model.decoder.layers.*.self_attention.linear_qkv.bias",
    target_key=(
        "model.layers.*.self_attn.q_proj.bias",
        "model.layers.*.self_attn.k_proj.bias",
        "model.layers.*.self_attn.v_proj.bias",
    ),
)
def _export_language_qkv_bias(ctx: io.TransformCTX, qkv_bias):
    # pylint: disable=C0115,C0116
    hf_config = ctx.target.config
    return export_qkv_bias(
        qkv_bias,
        head_num=hf_config.num_attention_heads,
        num_query_groups=hf_config.num_key_value_heads,
        heads_per_group=hf_config.num_attention_heads // hf_config.num_key_value_heads,
        head_size=hf_config.hidden_size // hf_config.num_attention_heads,
    )


@io.state_transform(
    source_key="vision_model.class_token",
    target_key="vision_model.embeddings.class_embedding",
)
def _export_cls_token(ctx: io.TransformCTX, cls_token):
    # pylint: disable=C0115,C0116
    return cls_token.squeeze()


@io.state_transform(
    source_key="language_model.decoder.layers.*.mlp.linear_fc1.weight",
    target_key=(
        "model.layers.*.mlp.gate_proj.weight",
        "model.layers.*.mlp.up_proj.weight",
    ),
)
def _export_linear_fc1(linear_fc1):
    # pylint: disable=C0115,C0116
    gate_proj, up_proj = torch.chunk(linear_fc1, 2, dim=0)
    return gate_proj, up_proj


@io.state_transform(
    source_key="vision_model.decoder.layers.*.mlp.linear_fc1.weight",
    target_key=(
        "visual.blocks.*.mlp.gate_proj.weight",
        "visual.blocks.*.mlp.up_proj.weight",
    ),
)
def _export_vision_linear_fc1_weight(vision_fc1_weight):
    # pylint: disable=C0115,C0116
    gate_proj, up_proj = torch.chunk(vision_fc1_weight, 2, dim=0)
    return gate_proj, up_proj


@io.state_transform(
    source_key="vision_model.decoder.layers.*.mlp.linear_fc1.bias",
    target_key=(
        "visual.blocks.*.mlp.gate_proj.bias",
        "visual.blocks.*.mlp.up_proj.bias",
    ),
)
def _export_vision_linear_fc1_bias(vision_fc1_bias):
    # pylint: disable=C0115,C0116
    gate_proj, up_proj = torch.chunk(vision_fc1_bias, 2, dim=0)
    return gate_proj, up_proj



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
from collections import defaultdict
from functools import partial
from typing import List, Optional

import regex as re
import torch
from lhotse import CutSet
from lhotse.cut import MixedCut
from omegaconf import DictConfig, OmegaConf
from torch import nn

from nemo.collections.asr.data.audio_to_text_lhotse_prompted import PromptedAudioToTextMiniBatch
from nemo.collections.asr.metrics.wer import WER
from nemo.core.classes import Serialization

__all__ = ['MultiTaskMetric']


# Helper functions for managing constraint criteria on metrics.
class ConstraintParser:
    """Boolean Parser class for constraint passing in config"""

    _primitives = None
    _booleans = None

    def parse_constraint(self, constraint: str):
        array = re.sub(r"([()])", r" \1 ", constraint).strip().split()  # Add space only for keywords.
        if not array:
            return self._no_constraint

        self._resolve_primitives(array)
        if len(array) == 1:
            return array[0]

        # Basic nested list parser. Starts from tail to aid readibility in subfunction.
        stack = []
        array = ["("] + array + [")"]
        while array:
            if (c := array.pop()) == "(":
                expr = []
                while stack:
                    if (e := stack.pop()) == ")":
                        if not (fnc := self._resolve_bools(expr)):
                            raise SyntaxError(f"Malformed subexpression find in constraint parsing: {fnc}")
                        stack.append(fnc)
                        break
                    expr.append(e)
            else:
                stack.append(c)
        if not (fnc := self._resolve_bools(stack)):
            raise SyntaxError(f"Parser cannot resolve constraint: {constraint}")
        return fnc

    @property
    def primitives(self):
        if self._primitives is None:
            self._primitives = {
                "==": operator.eq,
                "!=": operator.ne,
            }
        return self._primitives

    @property
    def booleans(self):
        if self._booleans is None:
            self._booleans = {
                "and": self._logical_and,
                "or": self._logical_or,
                "xor": self._logical_xor,
            }
        return self._booleans

    @staticmethod
    def _logical_not(expr, properties):
        if not expr:
            raise ValueError(f"Malformed subexpression find in 'not' constraint parsing: {expr}")
        return not expr(properties)

    @staticmethod
    def _logical_and(l_expr, r_expr, properties):
        if not (l_expr and r_expr):
            raise ValueError(f"Malformed subexpression find in 'and' constraint parsing: {l_expr} and {r_expr}")
        return l_expr(properties) and r_expr(properties)

    @staticmethod
    def _logical_or(l_expr, r_expr, properties):
        if not (l_expr and r_expr):
            raise ValueError(f"Malformed subexpression find in 'or' constraint parsing: {l_expr} or {r_expr}")
        return l_expr(properties) or r_expr(properties)

    @staticmethod
    def _logical_xor(l_expr, r_expr, properties):
        if not (l_expr and r_expr):
            raise ValueError(f"Malformed subexpression find in 'xor' constraint parsing: {l_expr} xor {r_expr}")
        return l_expr(properties) ^ r_expr(properties)

    @staticmethod
    def _no_constraint(properties):
        return True

    @staticmethod
    def _static_constraint(fnc, key, val, properties):
        return fnc(val, properties.get(key))

    @staticmethod
    def _compare_constraint(fnc, key1, key2, properties):
        return (
            (prop_val1 := properties.get(key1)) is not None
            and (prop_val2 := properties.get(key2)) is not None
            and fnc(prop_val1, prop_val2)
        )

    def _resolve_primitives(self, constraint):
        for idx, c in enumerate(constraint):
            for n, o in self.primitives.items():
                # Check if string is for value assertion or equivalency of values.
                entail, equal = fr'\.(\S+)\s*{n}\s*(\S+)', fr'\.(\S+)\s*{n}\s*\.(\S+)'
                match_entail, match_equal = re.match(entail, c), re.match(equal, c)
                if match_equal:
                    key1, key2 = match_equal.groups()
                    constraint[idx] = partial(self._compare_constraint, o, key1, key2)
                elif match_entail:
                    key1, val = match_entail.groups()
                    constraint[idx] = partial(self._static_constraint, o, key1, val)
                else:
                    pass

    def _resolve_bools(self, constraint: List[str]):
        idx = 0
        stack = []
        while idx < len(constraint):
            c = constraint[idx]
            if c == "not":
                c = partial(self._logical_not, constraint[idx + 1])
                idx += 1  # Skip so don't see the character again.
            stack.append(c)
            idx += 1

        constraint = stack
        for n, o in self.booleans.items():
            idx = 0
            stack = []
            while idx < len(constraint):
                c = constraint[idx]
                if c == n:
                    c = partial(o, stack.pop(), constraint[idx + 1])
                    idx += 1
                stack.append(c)
                idx += 1
            constraint = stack
        if len(constraint) > 1:  # More than one constraint, something went wrong.
            return None

        return constraint[0]


class MultiTaskMetric(Serialization):
    """
    Wrapper class for managing multiple metrics in multitask ASR/NLP models.

    This class enables conditional metric computation based on sample properties stored in Lhotse cuts.
    It's primarily designed for `EncDecMultiTaskModel` but can support any model with a prompt schema.

    Key Features:
        1. **Automatic Model Integration**: Instantiated metrics are automatically added as attributes
           to the parent model, enabling seamless integration with existing logging infrastructure.

        2. **Conditional Metric Updates**: Only samples meeting specific constraints are passed to
           each metric, avoiding inappropriate metric calculations (e.g., WER for translation tasks).

        3. **Flexible Constraint System**: Supports complex logical expressions for determining
           when metrics should be applied to samples.

        4. **Configuration Inheritance**: Global configuration parameters are automatically
           inherited by all metrics unless explicitly overridden.

    Args:
        model (nn.Module): Parent model that will receive metric instances as attributes.
                          Must have a `decoding` attribute for metrics that require decoding.
        cfg (DictConfig): Configuration dictionary containing metric definitions and constraints.

    Configuration Format:
        The configuration should follow this structure:

        ``'
        # Global parameters (inherited by all metrics unless overridden)
        log_predictions: true
        batch_dim_index: 0

        # Metric definitions
        metrics:
            wer:
                _target_: nemo.collections.asr.metrics.WER  # Metric class to instantiate
                constraint: ".task == transcribe"           # When to apply this metric
                use_cer: false                              # Metric-specific parameters
            bleu:
                _target_: nemo.collections.asr.metrics.BLEU
                constraint: ".task == translate"
                bleu_tokenizer: "13a"
                n_gram: 4

        ```

    Constraint Syntax:
        Constraints are evaluated against the `custom` dictionary of Lhotse cuts:

        - **Custom attribute Access**: `.task`, `.lang`, `.domain`
        - **Comparisons**: `==`, `!=`
        - **Logical Operations**: `and`, `or`, `not`, `xor`
        - **Property Comparisons**: `.source_lang == .target_lang`

        Examples:
        - `".task == transcribe"` - Apply to transcription tasks
        - `".task == translate and .source_lang != .target_lang"` - Cross-lingual translation
        - `"not .task == other"` - Apply to all tasks except 'other'
        - `".domain == medical or .domain == legal"` - Specific domains

    Usage Example:
        ```python
        # In model initialization
        if hasattr(cfg, 'multitask_metrics'):
            self.multitask_metrics = MultiTaskMetric(self, cfg.multitask_metrics)

        # During training/validation
        if hasattr(self, 'multitask_metrics'):
            metrics = self.multitask_metrics.eval(
                batch=batch,
                predictions=predictions,
                predictions_lengths=pred_lengths,
                predictions_mask=pred_mask,
                prefix="val",
                return_all_metrics=True
            )
            self.log_dict(metrics)
        ```

    Note:
        - Each metric receives the model's `decoding` instance for text decoding operations
        - Metrics are automatically instantiated for the parent model as attributes (e.g., `model.wer`, `model.bleu`)
        - Global configuration parameters are inherited unless explicitly overridden per metric
        - Metrics defined without 'constraint' keyword are called on every prediction sample
        - Empty batches (no samples matching constraints) are handled by child metrics.
    """

    def __init__(self, model: nn.Module, cfg: DictConfig):
        """
        Initialize MultiTaskMetric with model and configuration.

        Args:
            model (nn.Module): Parent model that will contain metric instances
            cfg (DictConfig): Configuration containing metric definitions
        """
        super().__init__()

        # Setup tracking dictionaries
        self._metric_dict, self._constr_dict = {}, {}
        cfg = OmegaConf.to_container(cfg)

        # Process each metric instance.
        parser = ConstraintParser()
        seen_types = set()
        for name, metric_cfg in cfg.pop("metrics").items():
            constraint = metric_cfg.pop(
                "constraint", ""
            )  # Empty string for no constraint value. Metric always calculated.

            # Inherit global configuration parameters
            for k, v in cfg.items():
                if k not in metric_cfg:  # do not override explicit metric values
                    metric_cfg[k] = v

            # Instantiates as instance of `model`. Avoids breaking behavior when other modules call specific metrics. (See `asr_model` for example.)
            metric_cfg["decoding"] = model.decoding  # For decoding reliant metrics like 'WER' or 'BLEU'
            metric = MultiTaskMetric.from_config_dict(metric_cfg)
            setattr(model, name, metric)

            # TODO: This is a from `asr_model` aggregation. To fix, update metric classes to support custom naming
            # and update `asr_model` `multi_{validation,test}_epoch_end` to support metric aggregation with custom names.
            metric_type = type(metric)
            if metric_type in seen_types:
                raise TypeError(
                    "MultiTaskMetric currently only supports one instance of each metric class. Please check your configs for duplicates values of `_target_` entry."
                )
            seen_types.add(metric_type)

            # Store metric and its constraint function
            self._metric_dict[name] = metric
            self._constr_dict[name] = parser.parse_constraint(constraint)

    # Performs full PyMetrics validation loop for all metrics.
    def eval(
        self,
        batch: PromptedAudioToTextMiniBatch,
        predictions: torch.Tensor,
        predictions_lengths: torch.Tensor,
        predictions_mask: torch.Tensor,
        return_all_metrics: Optional[bool] = False,
        prefix: Optional[str] = None,
    ):
        metric_dict = {}
        self.update(
            predictions=predictions,
            predictions_lengths=predictions_lengths,
            targets=batch.transcript,
            targets_lengths=batch.transcript_lens,
            predictions_mask=predictions_mask,
            input_ids=getattr(batch, "prompt", None),  # Allows for CTC and RNN-T support.
            cuts=batch.cuts,
        )
        metric_dict.update(
            self.compute(
                prefix=f"{prefix}_" if prefix else "",
                return_all_metrics=return_all_metrics,
            )
        )
        self.reset()
        return metric_dict

    def update(
        self,
        predictions: torch.Tensor,
        predictions_lengths: torch.Tensor,
        predictions_mask: torch.Tensor,
        targets: torch.Tensor,
        targets_lengths: torch.Tensor,
        input_ids: torch.Tensor,
        cuts: CutSet,
    ):

        # Update each metric with its filtered data
        cuts_split, idx_split = self._split_cuts(cuts)
        for name, metric in self._metric_dict.items():
            cuts_subset, indices = cuts_split[name], idx_split[name]
            # Update metric with filtered tensors
            metric.update(
                predictions=predictions[indices],
                predictions_lengths=predictions_lengths[indices],
                predictions_mask=predictions_mask[indices],
                targets=targets[indices],
                targets_lengths=targets_lengths[indices],
                input_ids=input_ids[indices],
                cuts=cuts_subset,
            )

    def compute(self, return_all_metrics=False, prefix=""):
        output_dict = {}

        for name, metric in self._metric_dict.items():
            # Handle WER metric's special return format
            # Custom name of metric used as suffix to allow custom naming.
            # TODO: Standardize WER to return dict like other metrics
            if type(metric) is WER:
                wer, wer_num, wer_denom = metric.compute()
                if return_all_metrics:
                    output_dict.update(
                        {
                            f"{prefix}wer": wer,
                            f"{prefix}wer_num": wer_num,
                            f"{prefix}wer_denom": wer_denom,
                        }
                    )
                else:
                    output_dict.update(
                        {
                            f"{prefix}wer": wer,
                        }
                    )
            else:
                # Standard metric compute (returns dict)
                output_dict.update(
                    metric.compute(
                        return_all_metrics=return_all_metrics,
                        prefix=prefix,
                    )
                )
        return output_dict

    def reset(self):
        {metric.reset() for name, metric in self._metric_dict.items()}

    def _split_cuts(self, cuts):
        """
        Split cuts based on metric constraints and return filtered subsets.

        This method evaluates each cut against all metric constraints and creates
        separate lists of cuts and indices for each metric.

        Args:
            cuts (CutSet): Input cuts containing sample metadata

        Returns:
            tuple: (cuts_splits, idx_splits) where:
                - cuts_splits (dict): Maps metric names to lists of matching cuts
                - idx_splits (dict): Maps metric names to lists of matching indices

        Note:
            - Handles both regular cuts and MixedCuts (uses first_non_padding_cut)
            - A single cut may match multiple metrics
            - Cuts not matching any constraints are ignored
        """
        cuts_splits, idx_splits = defaultdict(list), defaultdict(list)
        for idx, c in enumerate(cuts):
            c = c.first_non_padding_cut if isinstance(c, MixedCut) else c
            for metric, constr in self._constr_dict.items():
                if constr(c.custom):
                    cuts_splits[metric].append(c)
                    idx_splits[metric].append(idx)
        return cuts_splits, idx_splits



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import difflib
import os
from typing import List

import nemo_run as run
from lightning.pytorch.callbacks.callback import Callback
from nemo_run.core.serialization.yaml import YamlSerializer
from nemo_run.run.torchx_backend.packaging import _serialize

from nemo.collections.common.tokenizers.huggingface import AutoTokenizer
from nemo.collections.llm.gpt.data.squad import SquadDataModule
from nemo.collections.llm.gpt.model import GPTModel
from nemo.collections.llm.recipes.llama3_8b import MegatronCommOverlapCallback
from nemo.lightning.base import DEFAULT_NEMO_CACHE_HOME
from nemo.utils import logging

DEFAULT_NEMO_HOME = os.getenv('NEMO_HOME', DEFAULT_NEMO_CACHE_HOME)


def hf_tokenizer(model_name: str) -> run.Config[AutoTokenizer]:
    """
    HuggingFace tokenizer.

    Args:
        model_name (str): corresponds to HuggingFace-AutoTokenizer's 'pretrained_model_name_or_path' input argument.
                For more details please refer to-
                huggingface.co/docs/transformers/v4.47.1/en/model_doc/auto#transformers.AutoTokenizer
    """
    log_msg = [
        f"`AutoTokenizer` first searches for tokenizer files locally stored in {DEFAULT_NEMO_HOME}.",
        "(from env var `NEMO_HOME`- can be changed using '-nh/--nemo_home' CLI arg).",
        "If files are missing locally, `AutoTokenizer` will try downloading from HuggingFace. In this case-",
        "make sure env vars 'HF_HUB_OFFLINE':'0' and 'HF_TOKEN':'<token_value>' are set in your sbatch script.",
        "Both of these will be set automatically if you provide '-hf/--hf_token' CLI arg.",
    ]
    logging.warning(" ".join(log_msg))

    return run.Config(
        AutoTokenizer,
        pretrained_model_name=model_name,
        use_fast=True,
    )


def import_ckpt_experiment(executor: run.SlurmExecutor, model: run.Config[GPTModel], source: str):
    """
    Downloads/Acceses checkpoint to be used for fine-tuning. `import_ckpt` first tries find the nemo checkpoint in
    <NEMO_HOME>/models/. For eg: for llama3 8b, the path will look like- <NEMO_HOME>/models/meta-llama/Meta-Llama-3-8B
    If missing, tries to downloads at the same location from HuggingFace and converts it nemo format.

    Args:
        source (str): HuggingFace URL. For eg- hf://meta-llama/Meta-Llama-3-70B
    """
    from copy import deepcopy

    from nemo.collections.llm import import_ckpt

    import_executor = deepcopy(executor)
    import_executor.ntasks_per_node = 1
    import_executor.nodes = 1

    return run.Partial(import_ckpt, model=model, source=source, overwrite=False), import_executor, "import_ckpt_exp"


def get_nemo_home(nemo_home=None):
    """
    Get NEMO_HOME path. Checks for both nemo_home argument and NEMO_HOME environment variable.
    """
    arg_nemo_set = nemo_home is True
    env_nemo_set = "NEMO_HOME" in os.environ

    if arg_nemo_set and env_nemo_set:
        if os.environ["NEMO_HOME"] != nemo_home:
            logging.warning(f"Using nemo_home ({nemo_home}) instead of NEMO_HOME ({os.environ['NEMO_HOME']})")
        return nemo_home

    if arg_nemo_set:
        return nemo_home

    if env_nemo_set:
        return os.environ["NEMO_HOME"]

    raise ValueError("Neither -nh/--nemo_home argument nor NEMO_HOME environment variable is set")


def prepare_squad_dataset(model_name: str, seq_length: int = 2048, nemo_home=None):
    """Prepare the SQuAD dataset for fine-tuning.

    Args:
        model_name (str): The name of the model
        seq_length (int): The sequence length to use for packing. Defaults to 2048.
        nemo_home: Optional path to NEMO home directory set via args.nemo_home
    """
    from pathlib import Path

    from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer
    from nemo.collections.llm.gpt.data.packed_sequence import PackedSequenceSpecs
    from nemo.collections.llm.gpt.data.squad import SquadDataModule

    nemo_home_path = Path(get_nemo_home(nemo_home))
    dataset_root = nemo_home_path / "datasets" / "squad"
    dataset_root.mkdir(parents=True, exist_ok=True)

    tokenizer = AutoTokenizer(pretrained_model_name=model_name)

    # Configure SquadDataModule with packing specs
    datamodule = SquadDataModule(
        dataset_root=dataset_root,
        seq_length=seq_length,
        global_batch_size=8,
        micro_batch_size=1,
        packed_sequence_specs=PackedSequenceSpecs(packed_sequence_size=seq_length),
        tokenizer=tokenizer,
        force_redownload=True,
        delete_raw=False,
        seed=1234,
    )

    # This will generate both JSONL and packed .bin files
    datamodule.prepare_data()

    # Verify the output
    packed_dir = dataset_root / "packed" / model_name.replace("/", "--")
    print(f"Packed files should be in: {packed_dir}")
    if packed_dir.exists():
        print("Files found:", list(packed_dir.glob("*")))
    else:
        raise FileNotFoundError(f"Packed dataset dir not found at {packed_dir}. Dataset download failed")


def prepare_squad_dataset_experiment(
    executor: run.SlurmExecutor, model_name: str, seq_length: int = 2048, nemo_home=None
):
    """
    Downloads and prepares the SQuAD dataset for fine-tuning.
    """
    from copy import deepcopy

    dataset_executor = deepcopy(executor)
    dataset_executor.ntasks_per_node = 1
    dataset_executor.nodes = 1

    return (
        run.Partial(
            prepare_squad_dataset,
            model_name=model_name,
            seq_length=seq_length,
            nemo_home=nemo_home,
        ),
        dataset_executor,
        "prepare_squad_dataset_exp",
    )


def isfile_train_pack_metadata(hf_model_uri: str, data_config: run.Config[SquadDataModule]) -> bool:
    """
    This method is used for fine-tuning. It checks if packed train data for a partiular
    sequence length exists locally. This is needed to set data flag (force_redownload=True)
    which avoids experiment crash in case files are missing.
    """
    datasets_dir = os.getenv("NEMO_DATASETS_CACHE", os.path.join(DEFAULT_NEMO_HOME, "datasets"))
    model_dir = hf_model_uri.replace("/", "--")
    metadata_filename = f"{data_config.seq_length}_metadata.jsonl"

    train_pack_metadata_filepath = os.path.join(datasets_dir, "squad", "packed", model_dir, metadata_filename)

    return os.path.exists(train_pack_metadata_filepath) and os.path.isfile(train_pack_metadata_filepath)


def get_comm_overlap_callback_idx(callbacks: List[Callback]) -> int | None:
    """
    nemo.lightning.Trainer has a list of callbacks defined. This method identifies index of MegatronCommOverlapCallback
    from the list defined in recipes in nemo.collections.llm.recipes. The index is needed to override ddp communication
    params
    """
    if callbacks:  # default is None in lightning
        for idx, callback in enumerate(callbacks):
            if callback.__fn_or_cls__ == MegatronCommOverlapCallback:
                return idx
    return None


def dump_config_diff_from_base_recipe(
    base_recipe: str, new_recipe: str, output_dir: str, file_name: str = "config_diff.txt"
):
    """
    Dump the config diff from the base recipe.
    """
    base_recipe_config = _serialize(base_recipe, serializer_cls=YamlSerializer)
    new_recipe_config = _serialize(new_recipe, serializer_cls=YamlSerializer)
    diff = difflib.unified_diff(
        base_recipe_config.splitlines(keepends=True),
        new_recipe_config.splitlines(keepends=True),
        fromfile="base_recipe",
        tofile="new_recipe",
        lineterm="",
    )
    diff = "".join(diff)
    print("dumping config diff to ", os.path.join(output_dir, file_name))
    with open(os.path.join(output_dir, file_name), "w") as f:
        f.write(diff)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from loguru import logger
from pipecat.frames.frames import (
    BotStoppedSpeakingFrame,
    Frame,
    LLMFullResponseEndFrame,
    LLMFullResponseStartFrame,
    LLMTextFrame,
    TTSTextFrame,
)
from pipecat.observers.base_observer import FramePushed
from pipecat.processors.frameworks.rtvi import (
    RTVIBotLLMStartedMessage,
    RTVIBotLLMStoppedMessage,
    RTVIBotLLMTextMessage,
    RTVIBotTranscriptionMessage,
    RTVIBotTTSTextMessage,
)
from pipecat.processors.frameworks.rtvi import RTVIObserver as _RTVIObserver
from pipecat.processors.frameworks.rtvi import RTVIProcessor, RTVITextMessageData
from pipecat.transports.base_output import BaseOutputTransport

from nemo.agents.voice_agent.pipecat.utils.text.simple_text_aggregator import SimpleSegmentedTextAggregator


class RTVIObserver(_RTVIObserver):
    """
    An observer that processes RTVI frames and pushes them to the transport.
    """

    def __init__(
        self, rtvi: RTVIProcessor, text_aggregator: Optional[SimpleSegmentedTextAggregator] = None, *args, **kwargs
    ):
        super().__init__(rtvi, *args, **kwargs)
        self._text_aggregator = text_aggregator if text_aggregator else SimpleSegmentedTextAggregator("?!:.")

    async def on_push_frame(self, data: FramePushed):
        """Process a frame being pushed through the pipeline.

        Args:
            data: Frame push event data containing source, frame, direction, and timestamp.
        """
        src = data.source
        frame: Frame = data.frame

        if frame.id in self._frames_seen:
            return

        if not self._params.bot_llm_enabled:
            if isinstance(frame, LLMFullResponseStartFrame):
                await self.push_transport_message_urgent(RTVIBotLLMStartedMessage())
                self._frames_seen.add(frame.id)
            elif isinstance(frame, LLMFullResponseEndFrame):
                await self.push_transport_message_urgent(RTVIBotLLMStoppedMessage())
                self._frames_seen.add(frame.id)
            elif isinstance(frame, TTSTextFrame) and isinstance(src, BaseOutputTransport):
                message = RTVIBotTTSTextMessage(data=RTVITextMessageData(text=frame.text))
                await self.push_transport_message_urgent(message)
                await self._push_bot_transcription(frame.text)
                self._frames_seen.add(frame.id)
            elif isinstance(frame, BotStoppedSpeakingFrame):
                # The bot stopped speaking, flush the remaining text buffer
                await self._flush_text_buffer()
                self._frames_seen.add(frame.id)
            else:
                await super().on_push_frame(data)
        else:
            await super().on_push_frame(data)

    async def _flush_text_buffer(self):
        """Flush the text buffer."""
        text = self._text_aggregator.text
        if text.strip():
            await self._push_bot_transcription(text)
            self._text_aggregator.reset()

    async def _handle_llm_text_frame(self, frame: LLMTextFrame):
        """Handle LLM text output frames."""
        message = RTVIBotLLMTextMessage(data=RTVITextMessageData(text=frame.text))
        await self.push_transport_message_urgent(message)

        completed_text = await self._text_aggregator.aggregate(frame.text)
        if completed_text:
            await self._push_bot_transcription(completed_text)

    async def _push_bot_transcription(self, text: str):
        """Push accumulated bot transcription as a message."""
        if len(text.strip()) > 0:
            message = RTVIBotTranscriptionMessage(data=RTVITextMessageData(text=text))
            logger.debug(f"Pushing bot transcription: `{text}`")
            await self.push_transport_message_urgent(message)
            self._bot_transcription = ""



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tqdm import tqdm


class ProgressBar:
    """
    Base class for progress bars.
    """

    def __init__(self, value: float = 0.0, total: float = 1.0):
        """
        Initialize the ProgressBar.
        Args:
            value: (float) Initial value.
            total: (float) Total value. Must be greater than zero.
        """
        if total <= 0:
            raise ValueError("Total must be greater than zero.")
        if value < 0 or value > total:
            raise ValueError("Initial value must be between 0 and total.")

        self.value = value
        self.total = total
        self.start_value = value

    def restart(self) -> None:
        """Restart progress from the initial value."""
        self.value = self.start_value

    def increment(self, by: float) -> None:
        """
        Increase progress but do not exceed total.
        Args:
            by: (float) Amount to increment.
        """
        self.value = min(self.value + by, self.total)

    def update_bar(self, by: float) -> None:
        """
        Update progress and call update.
        Args:
            by: (float) Amount to increment.
        """
        self.increment(by)
        self.update()

    def finish(self) -> None:
        """Complete progress bar."""
        self.value = self.total
        self.update(True)

    def update(self, is_end: bool = False) -> None:
        """
        Abstract method for updating the progress bar.
        Args:
            is_end: (bool) Whether the progress bar is at the end.
        """
        raise NotImplementedError("Subclasses must implement update method.")


class TQDMProgressBar(ProgressBar):
    """TQDM progress bar wrapper."""

    def __init__(self, value: float = 0.0, total: float = 1.0):
        """
        Initialize the TQDMProgressBar.
        Args:
            value: (float) Initial value.
            total: (float) Total value.
        """
        super().__init__(value, total)
        self.bar = tqdm(total=self.total, bar_format='{l_bar}{bar}')
        self.prev_value = value

    def update(self, is_end: bool = False) -> None:
        """
        Update tqdm progress bar.
        Args:
            is_end: (bool) Whether the progress bar is at the end.
        """
        increment = self.value - self.prev_value
        if increment > 0:
            self.bar.update(increment)
        self.prev_value = self.value

        if is_end:
            self.bar.close()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, Dict, Generic, Optional, TypeVar, Union

import lightning.pytorch as pl
import torch
from lightning.fabric.plugins import CheckpointIO
from lightning.fabric.utilities.cloud_io import get_filesystem
from lightning.fabric.utilities.types import _PATH

try:
    from megatron.core.dist_checkpointing.serialization import (
        get_default_load_sharded_strategy,
        get_default_save_sharded_strategy,
    )
    from megatron.core.dist_checkpointing.strategies.base import SaveShardedStrategy
    from megatron.core.dist_checkpointing.strategies.fully_parallel import (
        FullyParallelLoadStrategyWrapper,
        FullyParallelSaveStrategyWrapper,
    )
    from megatron.core.dist_checkpointing.strategies.torch import TorchDistSaveShardedStrategy
    from megatron.core.parallel_state import get_data_parallel_group

    HAVE_MEGATRON_CORE = True

except (ImportError, ModuleNotFoundError):

    HAVE_MEGATRON_CORE = False

from torch import nn
from typing_extensions import Self, override

from nemo.lightning.ckpt_utils import WEIGHTS_PATH, ckpt_to_dir
from nemo.lightning.io.capture import IOProtocol
from nemo.lightning.io.mixin import IOMixin
from nemo.utils import logging

try:
    from nemo.utils.callbacks.dist_ckpt_io import AsyncCompatibleCheckpointIO
except ImportError:
    AsyncCompatibleCheckpointIO = CheckpointIO


LightningModuleT = TypeVar("LightningModuleT", bound=pl.LightningModule)
ModuleT = TypeVar("ModuleT", bound=nn.Module)


@dataclass
class TrainerContext(IOMixin, Generic[LightningModuleT]):
    """
    A context wrapper for a PyTorch Lightning Trainer and its associated model.

    This class ensures that both the trainer and its LightningModule extend `IOMixin`
    and provides additional context information.

    Attributes:
        model (LightningModuleT): The Lightning model associated with the trainer.
        trainer (pl.Trainer): The PyTorch Lightning trainer instance.
        extra (Dict[str, Any]): Additional context data, such as the `datamodule`, if available.
    """

    model: LightningModuleT
    trainer: pl.Trainer
    extra: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_trainer(cls, trainer: pl.Trainer) -> Self:
        """
        Creates a `TrainerContext` instance from a given `pl.Trainer`.

        Ensures that the trainer and its associated LightningModule support the `IOMixin` interface.

        Args:
            trainer (pl.Trainer): A PyTorch Lightning Trainer instance.

        Returns:
            TrainerContext: A new instance containing the trainer, model, and extra context.

        Raises:
            ValueError: If the trainer or its LightningModule does not extend `IOMixin`.
        """
        if not hasattr(trainer, "__io__"):
            raise ValueError(f"Trainer must be an instance of {IOProtocol}. Please use the Trainer from nemo.")
        if not hasattr(trainer.lightning_module, "__io__"):
            raise ValueError("LightningModule must extend IOMixin.")

        return cls(trainer=trainer, model=trainer.lightning_module, extra=cls.construct_extra(trainer))

    @classmethod
    def construct_extra(cls, trainer: pl.Trainer) -> Dict[str, Any]:
        """
        Constructs an `extra` dictionary containing additional relevant context.

        If the trainer has a `datamodule` that supports `IOMixin`, it will be added to `extra`.

        Args:
            trainer (pl.Trainer): A PyTorch Lightning Trainer instance.

        Returns:
            Dict[str, Any]: A dictionary containing extra context information.
        """
        extra = {}
        if hasattr(trainer, "datamodule") and hasattr(trainer.datamodule, "__io__"):
            extra["datamodule"] = trainer.datamodule.__io__

        return extra


def ckpt_to_weights_subdir(filepath: Union[str, Path], is_saving) -> Path:
    """Given an input checkpoint filepath, clean it using `ckpt_to_dir`
    and then return the weights subdirectory, if it exists."""
    from nemo.lightning.resume import AdapterPath

    filepath = ckpt_to_dir(filepath=filepath)
    base_dir = filepath
    assert not isinstance(base_dir, str)
    if base_dir.parts[-1] != WEIGHTS_PATH:
        maybe_base_dir = base_dir / WEIGHTS_PATH
        if maybe_base_dir.is_dir() or is_saving:
            base_dir = maybe_base_dir
            if isinstance(filepath, AdapterPath):
                base_dir.base_model_path = filepath.base_model_path
    # handle adapter paths
    if hasattr(base_dir, "base_model_path") and base_dir.base_model_path.parts[-1] != WEIGHTS_PATH:
        maybe_base_model_path = base_dir.base_model_path / WEIGHTS_PATH
        if maybe_base_model_path.is_dir() or is_saving:
            base_dir.base_model_path = base_dir.base_model_path / WEIGHTS_PATH
    if is_saving:
        assert base_dir.parts[-1] == WEIGHTS_PATH
        assert base_dir.parent == filepath
    return base_dir


class MegatronCheckpointIO(AsyncCompatibleCheckpointIO, IOMixin):
    """CheckpointIO that utilizes :func:`torch.save` and :func:`torch.load` to save and load checkpoints respectively,
    common for most use cases.

    .. warning::  This is an :ref:`experimental <versioning:Experimental API>` feature.

    """

    def __init__(
        self,
        save_ckpt_format: str = 'torch_dist',
        load_directly_on_device: bool = True,
        async_save: bool = False,
        torch_dist_multiproc: Optional[int] = None,
        assume_constant_structure: bool = False,
        parallel_save: bool = True,
        parallel_save_within_dp: bool = False,
        parallel_load: bool = False,
    ):
        self.save_ckpt_format = save_ckpt_format
        self.load_directly_on_device = load_directly_on_device
        self.async_save = async_save
        self.torch_dist_multiproc = torch_dist_multiproc
        self.assume_constant_structure = assume_constant_structure
        self.parallel_save = parallel_save
        self.parallel_save_within_dp = parallel_save_within_dp
        self.parallel_load = parallel_load

        self._save_sharded_strategy = None
        self.validated_consistency = False

    @override
    def save_checkpoint(
        self,
        checkpoint: Dict[str, Any],
        path: _PATH,
        storage_options: Optional[Any] = None,
    ) -> None:
        """Save model/training states as a checkpoint file through state-dump and file-write.

        Args:
            checkpoint: dict containing model and trainer state
            path: write-target path
            storage_options: if `storage_options` evaluates to True (e.g. non-empty dict)
                and `content_metadata` exists in
            content_metadata (dict, optional): metadata to identify the checkpoint content.
                Useful for framework specific versioning.
        Raises
        ------
            TypeError:
                If ``storage_options`` arg is passed in

        """
        from megatron.core import dist_checkpointing

        checkpoint_dir = ckpt_to_weights_subdir(path, is_saving=True)

        fs = get_filesystem(checkpoint_dir)
        fs.makedirs(checkpoint_dir, exist_ok=True)

        validate_sharding_integrity = not (self.validated_consistency and self.assume_constant_structure)
        self.validated_consistency = True

        rank = torch.distributed.get_rank()
        iteration = _get_iteration_from_checkpoint(checkpoint)
        start_time = time.time()
        async_save_request = dist_checkpointing.save(
            sharded_state_dict=checkpoint,
            checkpoint_dir=checkpoint_dir,
            sharded_strategy=self.save_sharded_strategy,
            validate_access_integrity=validate_sharding_integrity,
            async_sharded_save=self.async_save,
            content_metadata=(storage_options or {}).get('content_metadata'),
        )
        end_time = time.time()
        log_parts = (
            "Global Checkpoint Save",
            f"Rank: {rank}",
            f"Iteration: {iteration}" if iteration is not None else None,
            f"Start time: {start_time:.3f}s",
            f"Save duration: {end_time - start_time:.3f}s",
        )
        log_message = " : ".join(part for part in log_parts if part is not None)
        logging.info(log_message)

        def iter_finalize_fn():
            logging.info(f'Successfully saved checkpoint from iteration {int(iteration):7d} to {path}')

        if self.async_save:
            assert async_save_request is not None
            async_save_request.add_finalize_fn(iter_finalize_fn)

        return async_save_request

    @override
    def load_checkpoint(
        self,
        path: _PATH,
        sharded_state_dict=None,
        map_location: Optional[Callable] = None,
        strict: Optional['StrictHandling'] | bool = None,  # noqa: F821
    ) -> Dict[str, Any]:
        """Loads checkpoint using :func:`torch.load`, with additional handling for ``fsspec`` remote loading of files.

        Args:
            path: Path to checkpoint
            map_location: a function, :class:`torch.device`, string or a dict specifying how to remap storage
                locations.

        Returns: The loaded checkpoint.

        Raises
        ------
            FileNotFoundError: If ``path`` is not found by the ``fsspec`` filesystem

        """
        from megatron.core import dist_checkpointing
        from megatron.core.dist_checkpointing.validation import StrictHandling

        if map_location is not None:
            raise ValueError("`map_location` argument is not supported for `MegatronCheckpointIO.load_checkpoint`.")

        path = self._preprocess_checkpoint_load_path(path)

        if self.save_ckpt_format == 'zarr' and self.load_directly_on_device:
            from megatron.core.dist_checkpointing.strategies.tensorstore import TensorStoreLoadShardedStrategy

            sharded_strategy = TensorStoreLoadShardedStrategy(load_directly_on_device=True)
        else:
            sharded_strategy = None

        if self.parallel_load:
            if sharded_strategy is None:
                sharded_strategy = get_default_load_sharded_strategy(path)
            sharded_strategy = FullyParallelLoadStrategyWrapper(
                sharded_strategy, get_data_parallel_group(with_context_parallel=True)
            )

        if sharded_strategy is not None:
            logging.info(f'Using {sharded_strategy} dist-ckpt load strategy.')

        if isinstance(strict, bool):
            # For backward-compatibility reasons and a bug in MCore (strict check not applied to factories)
            # we must apply a simple strict check here.
            if not strict:
                sharded_state_dict = self.adjust_non_strict_load(path, sharded_state_dict)
            strict = StrictHandling.ASSUME_OK_UNEXPECTED if strict else StrictHandling.LOG_ALL
        if strict is None:
            # Default behavior
            strict = StrictHandling.ASSUME_OK_UNEXPECTED

        start_time = time.time()
        checkpoint = dist_checkpointing.load(
            sharded_state_dict=sharded_state_dict,
            checkpoint_dir=str(path),
            sharded_strategy=sharded_strategy,
            strict=strict,
        )
        checkpoint = _fix_tensors_device(checkpoint)
        end_time = time.time()
        duration = end_time - start_time
        logging.info(
            "Global Checkpoint Load : "
            f"Rank : {torch.distributed.get_rank()} : "
            f"Start time : {start_time:.3f}s : "
            f"Time spent in load_checkpoint: {duration:.3f}s"
        )
        return checkpoint

    @override
    def remove_checkpoint(self, path: _PATH) -> None:
        """Remove checkpoint file from the filesystem.

        Args:
            path: Path to checkpoint

        """
        fs = get_filesystem(path)
        if fs.exists(path):
            fs.rm(path, recursive=True)
            logging.debug(f"Removed checkpoint: {path}")

    def _determine_dist_ckpt_save_strategy(self):
        """Determine the saving strategy based on constructor args.

        Relies on the default MCore strategy unless extra PyT Distributed format arguments
        are passed in config or in case of a fully parallel save in which case
        a parallelization wrapper is applied.
        """
        if self.save_ckpt_format == 'zarr':
            logging.warning(
                '`zarr` distributed checkpoint backend is deprecated.'
                ' Distributed optimizer checkpoint saving might be extremely slow.'
                ' Please switch to PyTorch Distributed format (model.dist_ckpt_format=torch_dist).'
            )

        if self.async_save and self.save_ckpt_format != 'torch_dist':
            raise ValueError('Async dist-ckpt save supported only for torch_dist format')

        torch_dist_kwargs = {} if self.torch_dist_multiproc is None else dict(thread_count=self.torch_dist_multiproc)
        if self.save_ckpt_format == 'torch_dist' and torch_dist_kwargs:
            save_strategy = TorchDistSaveShardedStrategy(self.save_ckpt_format, 1, **torch_dist_kwargs)
        else:
            save_strategy = get_default_save_sharded_strategy(self.save_ckpt_format, 1)

        # MCore v0.8 introduces `use_cached_ckpt_structure` attribute
        if hasattr(save_strategy, 'use_cached_ckpt_structure'):
            save_strategy.use_cached_ckpt_structure = self.assume_constant_structure

        if self.parallel_save:
            parallelization_group = (
                get_data_parallel_group(with_context_parallel=True) if self.parallel_save_within_dp else None
            )
            save_strategy = FullyParallelSaveStrategyWrapper(
                save_strategy, parallelization_group, self.assume_constant_structure
            )

        logging.info(f'Using {save_strategy} dist-ckpt save strategy.')
        return save_strategy

    @property
    def save_sharded_strategy(self) -> 'SaveShardedStrategy':
        """
        initializes (if needed) the sharding strategy and returns its"""
        if self._save_sharded_strategy is None:
            self._save_sharded_strategy = self._determine_dist_ckpt_save_strategy()
        return self._save_sharded_strategy

    @staticmethod
    def _preprocess_checkpoint_load_path(path: _PATH):
        """Preprocess checkpoint path by checking if a directory exists and setting appropriate subdir.

        Args:
            path (_PATH): checkpoint path

        Returns:
            Path: preprocessed path that can be passed directly to `dist_checkpointing.load/save`

        Raises:
            FileNotFoundError: if path does not exist
            ValueError: if path is not a directory
        """
        # Try to read the checkpoint at `path`. If not exist, do not restore checkpoint.
        fs = get_filesystem(path)
        if not fs.exists(path):
            raise FileNotFoundError(f"Checkpoint file not found: {path}")
        if not fs.isdir(path):
            raise ValueError(f"Distributed checkpoints should be a directory. Found: {path}.")

        # Load from ckpt_path/weights (new format) if it exists
        path = ckpt_to_weights_subdir(path, is_saving=False)
        if hasattr(path, "base_model_path") and not path.base_model_path.exists():
            path.base_model_path = path.base_model_path.parent
        return path

    @staticmethod
    def load_content_metadata(path: Optional[_PATH] = None, preloaded_state_dict: Optional[dict] = None) -> dict:
        """Load content metadata stored in the checkpoint with `save_checkpoint(..., content_metadata=...)`.

        Args:
            path (_PATH, optional): checkpoint directory to load the content metadata from.
            preloaded_state_dict (StateDict, optional): if the state dict was already loaded,
                can be provided to avoid double load from storage

        Returns:
            dict: checkpoint content metadata
            None: in case there is no content metadata in the checkpoint
        """
        from megatron.core import dist_checkpointing

        if path is not None:
            path = MegatronCheckpointIO._preprocess_checkpoint_load_path(path)
        sharded_state_dict_metadata = dist_checkpointing.load_content_metadata(
            path, preloaded_state_dict=preloaded_state_dict
        )
        if sharded_state_dict_metadata is None:
            sharded_state_dict_metadata = {"distrib_optim_sharding_type": "fully_sharded_model_space"}
            logging.info(
                f"No content metadata in the checkpoint."
                f" Assuming backward compatibility metadata: {sharded_state_dict_metadata}"
            )
        else:
            logging.info(f'Loaded sharded_state_dict_metadata from checkpoint: {sharded_state_dict_metadata}')
        return sharded_state_dict_metadata

    def adjust_non_strict_load(self, path: _PATH, sharded_state_dict: Dict[str, Any]):
        """
        Adjusts the loading of a non-strict sharded checkpoint by filtering out missing keys.

        This function loads the checkpoint's metadata and removes any `ShardedBase` keys from
        `sharded_state_dict` that do not exist in the checkpoint. It also logs unexpected keys
        that were not found in the checkpoint.

        Args:
            path (_PATH): The path to the checkpoint.
            sharded_state_dict (Dict[str, Any]): The state dictionary containing sharded parameters.

        Returns:
            Dict[str, Any]: The adjusted state dictionary with missing keys removed.

        Notes:
            - Keys that exist in `sharded_state_dict` but are not found in the checkpoint metadata
            are considered "unexpected" and are logged.
            - Missing keys are not computed yet. To fully determine missing keys:
            1. Perform an `all_gather_object` operation on `loaded_keys`.
            2. Compute `missing_keys` as the difference between `ckpt_sharded_metadata.keys()`
                and `loaded_keys`.
        """
        from megatron.core import dist_checkpointing
        from megatron.core.dist_checkpointing.dict_utils import extract_matching_values
        from megatron.core.dist_checkpointing.mapping import ShardedBase

        ckpt_sharded_metadata = dist_checkpointing.load_tensors_metadata(path)
        loaded_keys = []
        unexpected_keys = []

        def should_remove_missing_sharded_base(x: Any):
            """
            Helper function to determine if a `ShardedBase` key should be removed.

            Args:
                x (Any): The object to check.

            Returns:
                bool: True if the key should be removed, False otherwise.
            """
            if isinstance(x, ShardedBase):
                if x.key in ckpt_sharded_metadata:
                    loaded_keys.append(x.key)
                    return False
                else:
                    unexpected_keys.append(x.key)
                    return True
            return False

        _, sharded_state_dict = extract_matching_values(sharded_state_dict, should_remove_missing_sharded_base)
        logging.info(f'The following keys are not in the checkpoint and will not be loaded: {unexpected_keys}')

        # TODO: compute missing_keys by:
        #  1. all_gather_object of loaded_keys
        #  2. missing_keys = ckpt_sharded_metadata.keys() - loaded_keys
        return sharded_state_dict


def _fix_tensors_device(ckpt: Dict) -> Dict:
    """Ensure checkpoint tensors are on the correct device."""
    assert torch.cuda.is_initialized(), (torch.cuda.is_available(), torch.cuda.is_initialized())
    cur_dev = torch.device("cuda", index=torch.cuda.current_device())
    from megatron.core.dist_checkpointing.dict_utils import dict_list_map_outplace

    def _fix_device(t):
        if isinstance(t, torch.Tensor) and t.is_cuda and t.device != cur_dev:
            t = t.to(cur_dev)
        return t

    return dict_list_map_outplace(_fix_device, ckpt)


def is_distributed_ckpt(path) -> bool:
    """Check if the given path corresponds to a distributed checkpoint directory.

    This function determines if the specified path is a directory that contains a distributed
    checkpoint by checking the directory's metadata.

    Args:
        path (Union[str, Path]): The path to check for being a distributed checkpoint.

    Returns
    -------
        bool: True if the path is a distributed checkpoint directory, False otherwise.

    """
    from megatron.core import dist_checkpointing

    checkpoint_dir = ckpt_to_dir(path)
    fs = get_filesystem(checkpoint_dir)
    return fs.isdir(checkpoint_dir) and dist_checkpointing.check_is_distributed_checkpoint(checkpoint_dir)


def _get_iteration_from_checkpoint(checkpoint: Dict[str, Any]) -> Optional[int]:
    return (
        checkpoint.get("loops", {})
        .get("fit_loop", {})
        .get("epoch_loop.batch_progress", {})
        .get("total", {})
        .get("completed", None)
    )



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging as _logging

from nemo.utils import logging as nemo_logger


def add_handlers_to_mcore_logger():
    """Add NeMo handlers to MCore loggers.

    MCore doesn't have and handlers for loggers (see
    https://docs.python.org/3/howto/logging-cookbook.html#adding-handlers-other-than-nullhandler-to-a-logger-in-a-library
    for a rationale). We have to add handlers explicitly.
    """
    mcore_logger = _logging.getLogger('megatron.core')
    for handler in nemo_logger._handlers.values():
        mcore_logger.addHandler(handler)
    mcore_logger.propagate = False
    mcore_logger.setLevel(nemo_logger._logger.level)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
from pathlib import Path

import torch
import transformers

from nemo.collections import llm
from nemo.collections.llm.utils import is_safe_repo


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--nemo-path", type=str, default="/root/.cache/nemo/models/models/llama_31_toy")
    parser.add_argument("--original-hf-path", type=str, default="models/llama_31_toy")
    parser.add_argument("--output-path", type=str, default="/tmp/output_hf")
    parser.add_argument("--add-model-name", action="store_true", default=False)
    parser.add_argument("--trust-remote-code", action="store_true", help="if trust_remote_code")
    parser.add_argument("--hf-target-class", type=str, default="AutoModelForCausalLM")
    parser.add_argument(
        "--allow-mismatch", nargs='+', default=[], help="List of parameter names to compare after casting to bfloat16"
    )
    return parser


if __name__ == '__main__':
    args = get_parser().parse_args()
    kwargs = {}
    if args.add_model_name:
        kwargs = {
            'target_model_name': args.original_hf_path,
        }
    try:
        llm.export_ckpt(
            path=Path(args.nemo_path),
            target='hf',
            output_path=Path(args.output_path),
            overwrite=True,
            trust_remote_code=args.trust_remote_code,
            **kwargs,
        )
    except TypeError:
        llm.export_ckpt(
            path=Path(args.nemo_path),
            target='hf',
            output_path=Path(args.output_path),
            overwrite=True,
            **kwargs,
        )

    hf_target_class = getattr(transformers, args.hf_target_class)
    original_hf = hf_target_class.from_pretrained(
        args.original_hf_path,
        trust_remote_code=is_safe_repo(
            hf_path=args.original_hf_path,
            trust_remote_code=args.trust_remote_code,
        ),
    )
    converted_hf = hf_target_class.from_pretrained(
        args.output_path,
        trust_remote_code=is_safe_repo(
            hf_path=args.original_hf_path,
            trust_remote_code=args.trust_remote_code,
        ),
    )

    for (name1, parameter1), (name2, parameter2) in zip(
        converted_hf.named_parameters(), original_hf.named_parameters()
    ):
        if any(k in name1 for k in args.allow_mismatch):
            param1_cmp = parameter1.bfloat16()
            param2_cmp = parameter2.bfloat16()
        else:
            param1_cmp = parameter1
            param2_cmp = parameter2
        assert torch.all(
            torch.isclose(param1_cmp, param2_cmp, atol=1e-3)
        ).item(), f'Parameter weight do not match for {name1}'

    print('All weights matched.')



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import os
import tempfile

import torch
import torch.distributed as dist

from nemo.utils import logging
from nemo.utils.get_rank import is_global_rank_zero

try:
    from megatron.core import parallel_state

    HAVE_MEGATRON_CORE = True
except (ImportError, ModuleNotFoundError):
    HAVE_MEGATRON_CORE = False


def initialize_distributed(args, backend='nccl'):
    """Initialize torch.distributed."""
    # Get local rank in case it is provided.
    local_rank = args.local_rank

    # Get rank and world size.
    rank = int(os.getenv('RANK', '0'))
    world_size = int(os.getenv("WORLD_SIZE", '1'))

    logging.info(
        f'Initializing torch.distributed with local_rank: {local_rank}, rank: {rank}, world_size: {world_size}'
    )

    # Set the device id.
    device = rank % torch.cuda.device_count()
    if local_rank is not None:
        device = local_rank
    torch.cuda.set_device(device)

    # Call the init process.
    init_method = 'tcp://'
    master_ip = os.getenv('MASTER_ADDR', 'localhost')
    master_port = os.getenv('MASTER_PORT', '6000')
    init_method += master_ip + ':' + master_port
    torch.distributed.init_process_group(backend=backend, world_size=world_size, rank=rank, init_method=init_method)
    return local_rank, rank, world_size


def gather_objects(partial_results_list, main_rank=None):
    """
    Collect objects (e.g., results) from all GPUs.
    Useful for inference over multiple GPUs with DDP.

    Use main_rank to specify which rank will be used to gather results.
    This allows to continue execution on the main_rank only after the gather.

    Args:
        partial_results_list: list of partial results from each GPU
        main_rank: rank of the main process to collect results from all GPUs (useful for collecting results in a target rank)


    Example:
        predictions = gather_objects(predictions,main_rank=0)
        # all but rank 0 will return None
        if predictions is None:
            return

        # from here only rank 0 should contiue
        pickle.dump(predictions, open(output_fname, "wb"))
    """
    # do not fail when DDP is not initialized
    if not parallel_state.is_initialized():
        return partial_results_list

    rank = parallel_state.get_data_parallel_rank()
    world_size = parallel_state.get_data_parallel_world_size()
    # return input when no DDP is used
    if world_size == 1:
        return partial_results_list

    gathered_results = [None for _ in range(world_size)]
    torch.distributed.all_gather_object(gathered_results, partial_results_list)

    # return None to non-main ranks
    if main_rank is not None:
        if rank != main_rank:
            return None

    # return collected results
    results_list = []
    for r in gathered_results:
        results_list.extend(r)

    return results_list


@contextlib.contextmanager
def temporary_directory():
    """Create a shared temporary directory across ranks in distributed setup.

    This function assumes that the distributed setup has been already
    correctly initialized. It is intended to be used only in single-node
    setup so that all ranks can access the directory created."""

    if is_global_rank_zero():
        tmp_dir = [tempfile.TemporaryDirectory()]
    else:
        tmp_dir = [None]
    dist.broadcast_object_list(tmp_dir)
    yield tmp_dir[0].name
    # We use barrier below to make sure that rank zero won't exit
    # and delete tmp_dir while other ranks may still use it
    dist.barrier()
    if is_global_rank_zero():
        tmp_dir[0].cleanup()


def webdataset_split_by_workers(src):
    """
    This is for latest webdataset>=0.2.6
    This function will make sure that each worker gets a different subset of the dataset.
    """
    # group = torch.distributed.group.WORLD
    # rank = torch.distributed.get_rank(group=group)
    # world_size = torch.distributed.get_world_size(group=group)
    worker_info = torch.utils.data.get_worker_info()
    num_workers = 1
    if worker_info is not None:
        worker = worker_info.id
        num_workers = worker_info.num_workers

    if num_workers > 1:
        yield from list(src)[worker::num_workers]
    else:
        yield from src



# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import librosa
import numpy as np
import torch
from torch import Tensor

from nemo.collections.asr.modules import AudioToMelSpectrogramPreprocessor
from nemo.collections.tts.parts.utils.tts_dataset_utils import get_audio_filepaths, normalize_volume, stack_tensors


class Featurizer(ABC):
    @abstractmethod
    def save(self, manifest_entry: Dict[str, Any], audio_dir: Path, feature_dir: Path, overwrite: bool = True) -> None:
        """
        Save feature value to disk for given manifest entry.

        Args:
            manifest_entry: Manifest entry dictionary.
            audio_dir: base directory where audio is stored.
            feature_dir: base directory where features will be stored.
            overwrite: whether to overwrite features if they already exist.
        """

    @abstractmethod
    def load(self, manifest_entry: Dict[str, Any], audio_dir: Path, feature_dir: Path) -> Dict[str, Tensor]:
        """
        Read saved feature value for given manifest entry.

        Args:
            manifest_entry: Manifest entry dictionary.
            audio_dir: base directory where audio is stored.
            feature_dir: base directory where features were stored by save().

        Returns:
            Dictionary of feature names to Tensors
        """

    @abstractmethod
    def collate_fn(self, train_batch: List[Dict[str, Tensor]]) -> Dict[str, Tensor]:
        """
        Combine list/batch of features into a feature dictionary.
        """


def _get_feature_filepath(
    manifest_entry: Dict[str, Any], audio_dir: Path, feature_dir: Path, feature_name: str
) -> Path:
    """
    Get the absolute path for the feature file corresponding to the input manifest entry

    Example: audio_filepath "<audio_dir>/speaker1/audio1.wav" becomes
        feature_filepath "<feature_dir>/<feature_name>/speaker1/audio1.npy"
    """
    _, audio_filepath_rel = get_audio_filepaths(manifest_entry=manifest_entry, audio_dir=audio_dir)
    feature_filepath = feature_dir / feature_name / audio_filepath_rel.with_suffix(".npy")
    return feature_filepath


def _features_exists(
    feature_names: List[Optional[str]],
    manifest_entry: Dict[str, Any],
    audio_dir: Path,
    feature_dir: Path,
) -> bool:
    for feature_name in feature_names:
        if feature_name is None:
            continue
        feature_filepath = _get_feature_filepath(
            manifest_entry=manifest_entry, audio_dir=audio_dir, feature_dir=feature_dir, feature_name=feature_name
        )
        if not feature_filepath.exists():
            return False
    return True


def _save_feature(
    feature_name: Optional[str],
    features: np.ndarray,
    manifest_entry: Dict[str, Any],
    audio_dir: Path,
    feature_dir: Path,
) -> None:
    """
    If feature_name is provided, save feature as .npy file.
    """
    if feature_name is None:
        return

    feature_filepath = _get_feature_filepath(
        manifest_entry=manifest_entry, audio_dir=audio_dir, feature_dir=feature_dir, feature_name=feature_name
    )
    feature_filepath.parent.mkdir(exist_ok=True, parents=True)
    np.save(file=str(feature_filepath), arr=features)


def _load_feature(
    feature_dict: Dict[str, Tensor],
    feature_name: Optional[str],
    manifest_entry: Dict[str, Any],
    audio_dir: Path,
    feature_dir: Path,
    indices: Optional[Tuple[int, int]] = None,
) -> None:
    """
    If feature_name is provided, load feature into feature_dict from .npy file.
    """
    if feature_name is None:
        return

    feature_filepath = _get_feature_filepath(
        manifest_entry=manifest_entry, audio_dir=audio_dir, feature_dir=feature_dir, feature_name=feature_name
    )
    feature_filepath = str(feature_filepath)
    if indices:
        feature_mmap = np.load(feature_filepath, mmap_mode='r')
        feature_array = feature_mmap[indices[0] : indices[1]]
        feature_array = np.copy(feature_array)
    else:
        feature_array = np.load(feature_filepath)
    feature_tensor = torch.from_numpy(feature_array)
    feature_dict[feature_name] = feature_tensor


def _get_frame_indices(manifest_entry: Dict[str, Any], sample_rate: int, hop_length: int) -> Optional[Tuple[int, int]]:
    if "offset" not in manifest_entry:
        return None

    offset = manifest_entry["offset"]
    duration = manifest_entry["duration"]
    start_i = librosa.core.time_to_frames(offset, sr=sample_rate, hop_length=hop_length)
    end_i = 1 + start_i + librosa.core.time_to_frames(duration, sr=sample_rate, hop_length=hop_length)
    return start_i, end_i


def _collate_feature(
    feature_dict: Dict[str, Tensor], feature_name: Optional[str], train_batch: List[Dict[str, Tensor]]
) -> None:
    if feature_name is None:
        return

    feature_tensors = []
    for example in train_batch:
        feature_tensor = example[feature_name]
        feature_tensors.append(feature_tensor)

    max_len = max([f.shape[0] for f in feature_tensors])
    stacked_features = stack_tensors(feature_tensors, max_lens=[max_len])
    feature_dict[feature_name] = stacked_features


class MelSpectrogramFeaturizer(Featurizer):
    def __init__(
        self,
        feature_name: str = "mel_spec",
        sample_rate: int = 22050,
        mel_dim: int = 80,
        win_length: int = 1024,
        hop_length: int = 256,
        lowfreq: int = 0,
        highfreq: int = 8000,
        log: bool = True,
        log_zero_guard_type: str = "add",
        log_zero_guard_value: float = 1.0,
        mel_norm: Optional[Union[str, int]] = None,
        volume_norm: bool = True,
    ) -> None:
        self.feature_name = feature_name
        self.sample_rate = sample_rate
        self.win_length = win_length
        self.hop_length = hop_length
        self.volume_norm = volume_norm

        self.preprocessor = AudioToMelSpectrogramPreprocessor(
            sample_rate=sample_rate,
            features=mel_dim,
            pad_to=1,
            n_window_size=win_length,
            n_window_stride=hop_length,
            window_size=False,
            window_stride=False,
            n_fft=win_length,
            lowfreq=lowfreq,
            highfreq=highfreq,
            mag_power=1.0,
            log=log,
            log_zero_guard_type=log_zero_guard_type,
            log_zero_guard_value=log_zero_guard_value,
            mel_norm=mel_norm,
            normalize=None,
            preemph=None,
            dither=0.0,
        )

    def compute_mel_spec(self, manifest_entry: Dict[str, Any], audio_dir: Path) -> np.ndarray:
        """
        Computes mel spectrogram for the input manifest entry.

        Args:
            manifest_entry: Manifest entry dictionary.
            audio_dir: base directory where audio is store

        Returns:
            [spec_dim, T_spec] float tensor containing spectrogram features.
        """
        audio_filepath_abs, _ = get_audio_filepaths(manifest_entry=manifest_entry, audio_dir=audio_dir)
        audio, _ = librosa.load(audio_filepath_abs, sr=self.sample_rate)

        if self.volume_norm:
            audio = normalize_volume(audio)

        # [1, T_audio]
        audio_tensor = torch.tensor(audio[np.newaxis, :], dtype=torch.float32)
        # [1]
        audio_len_tensor = torch.tensor([audio.shape[0]], dtype=torch.int32)

        # [1, spec_dim, T_spec]
        spec_tensor, _ = self.preprocessor(input_signal=audio_tensor, length=audio_len_tensor)
        # [spec_dim, T_spec]
        spec_tensor = spec_tensor.detach()[0]
        spec_array = spec_tensor.numpy()

        return spec_array

    def save(self, manifest_entry: Dict[str, Any], audio_dir: Path, feature_dir: Path, overwrite: bool = True) -> None:
        if not overwrite and _features_exists(
            feature_names=[self.feature_name],
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
        ):
            return

        spec = self.compute_mel_spec(manifest_entry=manifest_entry, audio_dir=audio_dir)
        _save_feature(
            feature_name=self.feature_name,
            features=spec,
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
        )

    def load(self, manifest_entry: Dict[str, Any], audio_dir: Path, feature_dir: Path) -> Dict[str, Tensor]:
        feature_dict = {}
        indices = _get_frame_indices(
            manifest_entry=manifest_entry, sample_rate=self.sample_rate, hop_length=self.hop_length
        )
        _load_feature(
            feature_dict=feature_dict,
            feature_name=self.feature_name,
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
            indices=indices,
        )
        return feature_dict

    def collate_fn(self, train_batch: List[Dict[str, Tensor]]) -> Dict[str, Tensor]:
        feature_dict = {}
        _collate_feature(feature_dict=feature_dict, feature_name=self.feature_name, train_batch=train_batch)
        return feature_dict


class EnergyFeaturizer(Featurizer):
    def __init__(self, spec_featurizer: MelSpectrogramFeaturizer, feature_name: str = "energy") -> None:
        self.feature_name = feature_name
        self.spec_featurizer = spec_featurizer

    def compute_energy(self, manifest_entry: Dict[str, Any], audio_dir: Path) -> np.ndarray:
        """
        Computes energy for the input manifest entry.

        Args:
            manifest_entry: Manifest entry dictionary.
            audio_dir: base directory where audio is store

        Returns:
            [T_spec] float tensor containing energy features.
        """
        # [spec_dim, T_spec]
        spec = self.spec_featurizer.compute_mel_spec(manifest_entry=manifest_entry, audio_dir=audio_dir)
        # [T_spec]
        energy = np.linalg.norm(spec, axis=0)
        return energy

    def save(self, manifest_entry: Dict[str, Any], audio_dir: Path, feature_dir: Path, overwrite: bool = True) -> None:
        if not overwrite and _features_exists(
            feature_names=[self.feature_name],
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
        ):
            return

        energy = self.compute_energy(manifest_entry=manifest_entry, audio_dir=audio_dir)
        _save_feature(
            feature_name=self.feature_name,
            features=energy,
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
        )

    def load(self, manifest_entry: Dict[str, Any], audio_dir: Path, feature_dir: Path) -> Dict[str, Tensor]:
        feature_dict = {}
        indices = _get_frame_indices(
            manifest_entry=manifest_entry,
            sample_rate=self.spec_featurizer.sample_rate,
            hop_length=self.spec_featurizer.hop_length,
        )
        _load_feature(
            feature_dict=feature_dict,
            feature_name=self.feature_name,
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
            indices=indices,
        )
        return feature_dict

    def collate_fn(self, train_batch: List[Dict[str, Tensor]]) -> Dict[str, Tensor]:
        feature_dict = {}
        _collate_feature(feature_dict=feature_dict, feature_name=self.feature_name, train_batch=train_batch)
        return feature_dict


class PitchFeaturizer(Featurizer):
    """
    Class for computing pitch features.

    Args:
        pitch_name: Optional directory name to save pitch features under.
            If None, then pitch will not be saved.
        voiced_mask_name: Optional directory name to save voiced mask under.
            If None, then voiced mask will not be saved.
        voiced_prob_name: Optional directory name to save voiced probabilities under.
            If None, then voiced probabilities will not be saved.
        sample_rate: Sample rate to use when loading audio.
        win_length: Audio frame length to use for pitch computation.
        hop_length: Audio hop length to use for pitch computation.
        pitch_fmin: Minimum pitch value to compute. Defaults to librosa.note_to_hz('C2') = 65.41 Hz.
        pitch_fmax: Maximum pitch value to compute. Defaults to librosa.note_to_hz('C7') = 2093.00 Hz.
            Setting this to a lower value will speed up computation, but may lose some pitch information.
        volume_norm: Whether to apply volume normalization to the audio.
        batch_seconds: Optional float, if provided then long audio files will have their pitch computed after
            splitting them into segments batch_seconds seconds long, to avoid running out of memory.
        batch_padding: If batch_seconds is provided, then this determines how many audio frames will be padded on
            both sides of each segment to ensure that the pitch values at the boundary are correct.
            If batch_seconds is not provided then this parameter is ignored.
    """

    def __init__(
        self,
        pitch_name: Optional[str] = "pitch",
        voiced_mask_name: Optional[str] = "voiced_mask",
        voiced_prob_name: Optional[str] = None,
        sample_rate: int = 22050,
        win_length: int = 1024,
        hop_length: int = 256,
        pitch_fmin: int = librosa.note_to_hz('C2'),
        pitch_fmax: int = librosa.note_to_hz('C7'),
        volume_norm: bool = True,
        batch_seconds: Optional[float] = 30.0,
        batch_padding: int = 10,
    ) -> None:
        self.pitch_name = pitch_name
        self.voiced_mask_name = voiced_mask_name
        self.voiced_prob_name = voiced_prob_name
        self.sample_rate = sample_rate
        self.win_length = win_length
        self.hop_length = hop_length
        self.volume_norm = volume_norm
        self.pitch_fmin = pitch_fmin
        self.pitch_fmax = pitch_fmax
        if batch_seconds:
            assert batch_padding is not None
            # Round sample size up to a multiple of hop_length
            batch_samples = int(batch_seconds * sample_rate)
            self.batch_frames = int(math.ceil(batch_samples / self.hop_length))
            self.batch_samples = self.hop_length * self.batch_frames
            self.batch_padding_frames = batch_padding
            self.batch_padding_samples = self.hop_length * self.batch_padding_frames
        else:
            self.batch_samples = None
            self.batch_padding = None

    def compute_pitch(
        self, manifest_entry: Dict[str, Any], audio_dir: Path
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes pitch and optional voiced mask for the input manifest entry.

        Args:
            manifest_entry: Manifest entry dictionary.
            audio_dir: base directory where audio is store

        Returns:
            pitch: [T_spec] float tensor containing pitch for each audio frame.
            voiced_mask: [T_spec] bool tensor indicating whether each audio frame is voiced.
            voiced_prob: [T_spec] float array with [0, 1] probability that each audio frame is voiced.
        """
        audio_filepath_abs, _ = get_audio_filepaths(manifest_entry=manifest_entry, audio_dir=audio_dir)
        audio, _ = librosa.load(audio_filepath_abs, sr=self.sample_rate)

        if self.volume_norm:
            audio = normalize_volume(audio)

        if not self.batch_samples or audio.shape[0] < self.batch_samples:
            pitch, voiced_mask, voiced_prob = librosa.pyin(
                audio,
                fmin=self.pitch_fmin,
                fmax=self.pitch_fmax,
                sr=self.sample_rate,
                frame_length=self.win_length,
                hop_length=self.hop_length,
                fill_na=0.0,
            )
        else:
            num_chunks = int(np.ceil(audio.shape[0] / self.batch_samples))
            pitch_list = []
            voiced_mask_list = []
            voiced_prob_list = []

            for i in range(num_chunks):
                start_i = i * self.batch_samples
                end_i = (i + 1) * self.batch_samples

                if i != 0:
                    # Pad beginning with additional frames
                    start_i -= self.batch_padding_samples
                if i != (num_chunks - 1):
                    # Pad end with additional frames
                    end_i += self.batch_padding_samples

                audio_chunk = audio[start_i:end_i]
                pitch_i, voiced_mask_i, voiced_prob_i = librosa.pyin(
                    audio_chunk,
                    fmin=self.pitch_fmin,
                    fmax=self.pitch_fmax,
                    sr=self.sample_rate,
                    frame_length=self.win_length,
                    hop_length=self.hop_length,
                    fill_na=0.0,
                )
                # Remove padded frames
                if i != 0:
                    pitch_i = pitch_i[self.batch_padding_frames :]
                    voiced_mask_i = voiced_mask_i[self.batch_padding_frames :]
                    voiced_prob_i = voiced_prob_i[self.batch_padding_frames :]
                if i != (num_chunks - 1):
                    pitch_i = pitch_i[: self.batch_frames]
                    voiced_mask_i = voiced_mask_i[: self.batch_frames]
                    voiced_prob_i = voiced_prob_i[: self.batch_frames]

                pitch_list.append(pitch_i)
                voiced_mask_list.append(voiced_mask_i)
                voiced_prob_list.append(voiced_prob_i)

            pitch = np.concatenate(pitch_list, axis=0)
            voiced_mask = np.concatenate(voiced_mask_list, axis=0)
            voiced_prob = np.concatenate(voiced_prob_list, axis=0)

        pitch = pitch.astype(np.float32)
        voiced_prob = voiced_prob.astype(np.float32)

        return pitch, voiced_mask, voiced_prob

    def save(self, manifest_entry: Dict[str, Any], audio_dir: Path, feature_dir: Path, overwrite: bool = True) -> None:
        if not overwrite and _features_exists(
            feature_names=[self.pitch_name, self.voiced_mask_name, self.voiced_prob_name],
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
        ):
            return

        pitch, voiced_mask, voiced_prob = self.compute_pitch(manifest_entry=manifest_entry, audio_dir=audio_dir)
        _save_feature(
            feature_name=self.pitch_name,
            features=pitch,
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
        )
        _save_feature(
            feature_name=self.voiced_mask_name,
            features=voiced_mask,
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
        )
        _save_feature(
            feature_name=self.voiced_prob_name,
            features=voiced_prob,
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
        )

    def load(self, manifest_entry: Dict[str, Any], audio_dir: Path, feature_dir: Path) -> Dict[str, Tensor]:
        feature_dict = {}
        indices = _get_frame_indices(
            manifest_entry=manifest_entry, sample_rate=self.sample_rate, hop_length=self.hop_length
        )
        _load_feature(
            feature_dict=feature_dict,
            feature_name=self.pitch_name,
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
            indices=indices,
        )
        _load_feature(
            feature_dict=feature_dict,
            feature_name=self.voiced_mask_name,
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
            indices=indices,
        )
        _load_feature(
            feature_dict=feature_dict,
            feature_name=self.voiced_prob_name,
            manifest_entry=manifest_entry,
            audio_dir=audio_dir,
            feature_dir=feature_dir,
            indices=indices,
        )
        return feature_dict

    def collate_fn(self, train_batch: List[Dict[str, Tensor]]) -> Dict[str, Tensor]:
        feature_dict = {}
        _collate_feature(feature_dict=feature_dict, feature_name=self.pitch_name, train_batch=train_batch)
        _collate_feature(feature_dict=feature_dict, feature_name=self.voiced_mask_name, train_batch=train_batch)
        _collate_feature(feature_dict=feature_dict, feature_name=self.voiced_prob_name, train_batch=train_batch)
        return feature_dict



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
import shutil
import tempfile

import pytest
import torch
from lhotse import CutSet, MonoCut
from lhotse.testing.dummies import DummyManifest
from omegaconf import DictConfig

from nemo.collections.asr.data import audio_to_text
from nemo.collections.asr.data.audio_to_text_lhotse import LhotseSpeechToTextBpeDataset
from nemo.collections.asr.models import configs
from nemo.collections.asr.models.ctc_bpe_models import EncDecCTCModelBPE
from nemo.collections.asr.parts.submodules import ctc_beam_decoding as beam_decode
from nemo.collections.asr.parts.submodules.ctc_decoding import CTCBPEDecoding, CTCBPEDecodingConfig
from nemo.collections.asr.parts.utils.rnnt_utils import Hypothesis
from nemo.collections.common import tokenizers
from nemo.utils.config_utils import assert_dataclass_signature_match


@pytest.fixture()
def asr_model(test_data_dir):
    preprocessor = {'_target_': 'nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor'}
    encoder = {
        '_target_': 'nemo.collections.asr.modules.ConvASREncoder',
        'feat_in': 64,
        'activation': 'relu',
        'conv_mask': True,
        'jasper': [
            {
                'filters': 1024,
                'repeat': 1,
                'kernel': [1],
                'stride': [1],
                'dilation': [1],
                'dropout': 0.0,
                'residual': False,
                'separable': True,
                'se': True,
                'se_context_size': -1,
            }
        ],
    }

    decoder = {
        '_target_': 'nemo.collections.asr.modules.ConvASRDecoder',
        'feat_in': 1024,
        'num_classes': -1,
        'vocabulary': None,
    }

    tokenizer = {'dir': os.path.join(test_data_dir, "asr", "tokenizers", "an4_wpe_128"), 'type': 'wpe'}

    modelConfig = DictConfig(
        {
            'preprocessor': DictConfig(preprocessor),
            'encoder': DictConfig(encoder),
            'decoder': DictConfig(decoder),
            'tokenizer': DictConfig(tokenizer),
        }
    )

    model_instance = EncDecCTCModelBPE(cfg=modelConfig)
    return model_instance


class TestEncDecCTCModel:
    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_constructor(self, asr_model):
        asr_model.train()
        # TODO: make proper config and assert correct number of weights
        # Check to/from config_dict:
        confdict = asr_model.to_config_dict()
        instance2 = EncDecCTCModelBPE.from_config_dict(confdict)
        assert isinstance(instance2, EncDecCTCModelBPE)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_forward(self, asr_model):
        asr_model = asr_model.eval()

        asr_model.preprocessor.featurizer.dither = 0.0
        asr_model.preprocessor.featurizer.pad_to = 0

        input_signal = torch.randn(size=(4, 512))
        length = torch.randint(low=321, high=500, size=[4])

        with torch.no_grad():
            # batch size 1
            logprobs_instance = []
            for i in range(input_signal.size(0)):
                logprobs_ins, _, _ = asr_model.forward(
                    input_signal=input_signal[i : i + 1], input_signal_length=length[i : i + 1]
                )
                logprobs_instance.append(logprobs_ins)
                print(len(logprobs_ins))
            logprobs_instance = torch.cat(logprobs_instance, 0)

            # batch size 4
            logprobs_batch, _, _ = asr_model.forward(input_signal=input_signal, input_signal_length=length)

        assert logprobs_instance.shape == logprobs_batch.shape
        diff = torch.mean(torch.abs(logprobs_instance - logprobs_batch))
        assert diff <= 1e-6
        diff = torch.max(torch.abs(logprobs_instance - logprobs_batch))
        assert diff <= 1e-6

    @pytest.mark.unit
    def test_predict_step(self, asr_model):
        asr_model = asr_model.eval()
        cuts = DummyManifest(CutSet, begin_id=0, end_id=1, with_data=True)
        dataset = LhotseSpeechToTextBpeDataset(tokenizer=asr_model.tokenizer, return_cuts=True)
        batch = dataset[cuts]
        outputs = asr_model.predict_step(batch, 0)
        assert len(outputs) == 1
        assert len(outputs[0]) == 2
        assert isinstance(outputs[0][0], MonoCut)
        assert isinstance(outputs[0][1], Hypothesis)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_save_restore_artifact(self, asr_model):
        with tempfile.TemporaryDirectory() as tmpdir:
            save_path = os.path.join(tmpdir, 'ctc_bpe.nemo')
            asr_model.train()
            asr_model.save_to(save_path)

            new_model = EncDecCTCModelBPE.restore_from(save_path)
            assert isinstance(new_model, type(asr_model))
            assert new_model.vocab_path.endswith('_vocab.txt')

            assert len(new_model.tokenizer.tokenizer.get_vocab()) == 128

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_save_restore_artifact_spe(self, asr_model, test_data_dir):
        with tempfile.TemporaryDirectory() as tmpdir:
            tokenizer_dir = os.path.join(test_data_dir, "asr", "tokenizers", "an4_spe_128")
            asr_model.change_vocabulary(new_tokenizer_dir=tokenizer_dir, new_tokenizer_type='bpe')

            save_path = os.path.join(tmpdir, 'ctc_bpe.nemo')
            asr_model.train()
            asr_model.save_to(save_path)

            new_model = EncDecCTCModelBPE.restore_from(save_path)
            assert isinstance(new_model, type(asr_model))
            assert isinstance(new_model.tokenizer, tokenizers.SentencePieceTokenizer)
            assert new_model.model_path.endswith('_tokenizer.model')
            assert new_model.vocab_path.endswith('_vocab.txt')
            assert new_model.spe_vocab_path.endswith('_tokenizer.vocab')

            assert new_model.tokenizer.tokenizer.vocab_size == 128
            assert len(new_model.tokenizer.tokenizer.get_vocab()) == 128

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_save_restore_artifact_agg(self, asr_model, test_data_dir):
        tokenizer_dir = os.path.join(test_data_dir, "asr", "tokenizers", "an4_spe_128")
        tok_en = {"dir": tokenizer_dir, "type": "wpe"}
        # the below is really an english tokenizer but we pretend it is spanish
        tok_es = {"dir": tokenizer_dir, "type": "wpe"}
        tcfg = DictConfig({"type": "agg", "langs": {"en": tok_en, "es": tok_es}})
        with tempfile.TemporaryDirectory() as tmpdir:
            asr_model.change_vocabulary(new_tokenizer_dir=tcfg, new_tokenizer_type="agg")

            save_path = os.path.join(tmpdir, "ctc_agg.nemo")
            asr_model.train()
            asr_model.save_to(save_path)

            new_model = EncDecCTCModelBPE.restore_from(save_path)
            assert isinstance(new_model, type(asr_model))
            assert isinstance(new_model.tokenizer, tokenizers.AggregateTokenizer)

            # should be double
            assert new_model.tokenizer.tokenizer.vocab_size == 254
            assert len(new_model.tokenizer.tokenizer.get_vocab()) == 254

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_vocab_change(self, test_data_dir, asr_model):
        old_vocab = copy.deepcopy(asr_model.decoder.vocabulary)

        with tempfile.TemporaryDirectory() as save_dir:
            save_path = os.path.join(save_dir, 'temp.nemo')

            with tempfile.TemporaryDirectory() as tmpdir:
                old_tmpdir_path = tmpdir

                old_tokenizer_dir = os.path.join(test_data_dir, "asr", "tokenizers", "an4_wpe_128", 'vocab.txt')
                new_tokenizer_dir = os.path.join(tmpdir, 'tokenizer')

                os.makedirs(new_tokenizer_dir, exist_ok=True)
                shutil.copy2(old_tokenizer_dir, new_tokenizer_dir)

                nw1 = asr_model.num_weights
                asr_model.change_vocabulary(new_tokenizer_dir=new_tokenizer_dir, new_tokenizer_type='wpe')
                # No change
                assert nw1 == asr_model.num_weights

                with open(os.path.join(new_tokenizer_dir, 'vocab.txt'), 'a+') as f:
                    f.write("!\n")
                    f.write('$\n')
                    f.write('@\n')

                asr_model.change_vocabulary(new_tokenizer_dir=new_tokenizer_dir, new_tokenizer_type='wpe')
                # fully connected + bias
                assert asr_model.num_weights == nw1 + 3 * (asr_model.decoder._feat_in + 1)

                new_vocab = copy.deepcopy(asr_model.decoder.vocabulary)
                assert len(old_vocab) != len(new_vocab)

                # save the model (after change of vocabulary)
                asr_model.save_to(save_path)
                assert os.path.exists(save_path)
                # delete copied version of the vocabulary from nested tmpdir (by scope)

            # assert copied vocab no longer exists
            assert not os.path.exists(os.path.join(old_tmpdir_path, 'tokenizer', 'vocab.txt'))

            # make a copy of the tokenizer before renaming
            try:
                os.rename(old_tokenizer_dir, old_tokenizer_dir + '.bkp')
                assert not os.path.exists(old_tokenizer_dir)

                # restore model from .nemo
                asr_model2 = EncDecCTCModelBPE.restore_from(save_path)
                assert isinstance(asr_model2, EncDecCTCModelBPE)

                # Check if vocabulary size is same
                assert asr_model.tokenizer.tokenizer.vocab_size == asr_model2.tokenizer.tokenizer.vocab_size

                # Make a copy of the tokenizer
                new_tokenizer_dir = os.path.join(save_dir, 'tokenizer')

                os.makedirs(new_tokenizer_dir, exist_ok=True)
                new_tokenizer_path = os.path.join(new_tokenizer_dir, 'vocab.txt')
                with open(new_tokenizer_path, 'w') as f:
                    for v in asr_model2.tokenizer.tokenizer.get_vocab():
                        f.write(f"{v}\n")

                    # Add some new tokens too
                    f.write("^\n")
                    f.write("^^\n")
                    f.write("^^^\n")

                assert os.path.exists(new_tokenizer_path)

                # change vocabulary
                asr_model2.change_vocabulary(new_tokenizer_dir, new_tokenizer_type='wpe')
                assert asr_model.tokenizer.vocab_size != asr_model2.tokenizer.vocab_size

                new_save_path = os.path.join(save_dir, 'temp2.nemo')
                asr_model2.save_to(new_save_path)

                asr_model3 = EncDecCTCModelBPE.restore_from(new_save_path)
                assert isinstance(asr_model3, EncDecCTCModelBPE)

                # Check if vocabulary size is same
                assert asr_model2.tokenizer.tokenizer.vocab_size == asr_model3.tokenizer.tokenizer.vocab_size
                assert asr_model2.vocab_path != asr_model3.vocab_path

                # Model PT level checks
                assert len(asr_model2.artifacts) == 1

            finally:
                os.rename(old_tokenizer_dir + '.bkp', old_tokenizer_dir)

    @pytest.mark.unit
    def test_decoding_change(self, asr_model):
        assert asr_model.decoding is not None
        assert isinstance(asr_model.decoding, CTCBPEDecoding)
        assert asr_model.decoding.cfg.strategy == "greedy_batch"
        assert asr_model.decoding.preserve_alignments is False
        assert asr_model.decoding.compute_timestamps is False

        cfg = CTCBPEDecodingConfig(preserve_alignments=True, compute_timestamps=True)
        asr_model.change_decoding_strategy(cfg)

        assert asr_model.decoding.preserve_alignments is True
        assert asr_model.decoding.compute_timestamps is True

        new_strategy = DictConfig({})
        new_strategy.strategy = 'beam'
        new_strategy.beam = DictConfig({'beam_size': 1})
        asr_model.change_decoding_strategy(decoding_cfg=new_strategy)
        assert isinstance(asr_model.decoding.decoding, beam_decode.BeamCTCInfer)
        assert asr_model.decoding.decoding.search_type == "default"

        new_strategy = DictConfig({})
        new_strategy.strategy = 'pyctcdecode'
        new_strategy.beam = DictConfig({'beam_size': 1})
        asr_model.change_decoding_strategy(decoding_cfg=new_strategy)
        assert isinstance(asr_model.decoding.decoding, beam_decode.BeamCTCInfer)
        assert asr_model.decoding.decoding.search_type == "pyctcdecode"

        new_strategy = DictConfig({})
        new_strategy.strategy = 'flashlight'
        new_strategy.beam = DictConfig({'beam_size': 1})
        asr_model.change_decoding_strategy(decoding_cfg=new_strategy)
        assert isinstance(asr_model.decoding.decoding, beam_decode.BeamCTCInfer)
        assert asr_model.decoding.decoding.search_type == "flashlight"

        new_strategy = DictConfig({})
        new_strategy.strategy = 'wfst'
        new_strategy.beam = DictConfig({'beam_size': 1})
        asr_model.change_decoding_strategy(decoding_cfg=new_strategy)
        assert isinstance(asr_model.decoding.decoding, beam_decode.WfstCTCInfer)
        assert asr_model.decoding.decoding.search_type == "riva"

    @pytest.mark.unit
    def test_ASRDatasetConfig_for_AudioToBPEDataset(self):
        # ignore some additional arguments as dataclass is generic
        IGNORE_ARGS = [
            'is_tarred',
            'num_workers',
            'batch_size',
            'tarred_audio_filepaths',
            'shuffle',
            'pin_memory',
            'drop_last',
            'tarred_shard_strategy',
            'shard_manifests',
            'shuffle_n',
            'parser',
            'normalize',
            'unk_index',
            'pad_id',
            'bos_id',
            'eos_id',
            'blank_index',
            'bucketing_batch_size',
            'bucketing_strategy',
            'bucketing_weights',
            'channel_selector',
            'use_lhotse',
            'tarred_random_access',
            'use_bucketing',
            'batch_duration',
            'quadratic_duration',
            'bucket_batch_size',
            'bucket_duration_bins',
            'num_buckets',
            'pin_memory',
        ]

        REMAP_ARGS = {'trim_silence': 'trim', 'labels': 'tokenizer'}

        result = assert_dataclass_signature_match(
            audio_to_text.AudioToBPEDataset,
            configs.ASRDatasetConfig,
            ignore_args=IGNORE_ARGS,
            remap_args=REMAP_ARGS,
        )
        signatures_match, cls_subset, dataclass_subset = result

        assert signatures_match
        assert cls_subset is None
        assert dataclass_subset is None

    @pytest.mark.unit
    def test_ASRDatasetConfig_for_TarredAudioToBPEDataset(self):
        # ignore some additional arguments as dataclass is generic
        IGNORE_ARGS = [
            'is_tarred',
            'num_workers',
            'batch_size',
            'shuffle',
            'pin_memory',
            'drop_last',
            'parser',
            'normalize',
            'unk_index',
            'pad_id',
            'bos_id',
            'eos_id',
            'blank_index',
            'global_rank',
            'world_size',
            'bucketing_batch_size',
            'bucketing_strategy',
            'bucketing_weights',
            'max_utts',
            'use_lhotse',
            'tarred_random_access',
            'use_bucketing',
            'batch_duration',
            'quadratic_duration',
            'bucket_batch_size',
            'bucket_duration_bins',
            'num_buckets',
            'pin_memory',
        ]

        REMAP_ARGS = {
            'trim_silence': 'trim',
            'tarred_audio_filepaths': 'audio_tar_filepaths',
            'tarred_shard_strategy': 'shard_strategy',
            'shuffle_n': 'shuffle',
            'labels': 'tokenizer',
        }

        result = assert_dataclass_signature_match(
            audio_to_text.TarredAudioToBPEDataset,
            configs.ASRDatasetConfig,
            ignore_args=IGNORE_ARGS,
            remap_args=REMAP_ARGS,
        )
        signatures_match, cls_subset, dataclass_subset = result

        assert signatures_match
        assert cls_subset is None
        assert dataclass_subset is None



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch

from nemo.core import NeuralModule
from nemo.core.classes.mixins import adapter_mixin_strategies, adapter_mixins
from nemo.core.classes.mixins.adapter_mixins import AdapterModuleMixin


class DefaultModule(NeuralModule):
    def __init__(self):
        super().__init__()

        self.fc = torch.nn.Linear(50, 50)
        self.bn = torch.nn.BatchNorm1d(50)

    def forward(self, x):
        x = self.fc(x)
        x = self.bn(x)
        out = x
        return out

    def num_params(self):
        num: int = 0
        for p in self.parameters():
            if p.requires_grad:
                num += p.numel()
        return num


class DefaultModuleAdapter(DefaultModule, AdapterModuleMixin):
    def forward(self, x):
        x = super(DefaultModuleAdapter, self).forward(x)

        if self.is_adapter_available():
            # For testing purposes, cache the adapter names
            self._adapter_names = self.get_enabled_adapters()
            # call forward over model adapters, summing them up
            x = self.forward_enabled_adapters(x)

        return x


def get_adapter_cfg(in_features=50, dim=100, norm_pos='pre'):
    cfg = {
        '_target_': 'nemo.collections.common.parts.adapter_modules.LinearAdapter',
        'in_features': in_features,
        'dim': dim,
        'norm_position': norm_pos,
    }
    return cfg


def get_classpath(cls):
    return f'{cls.__module__}.{cls.__name__}'


if adapter_mixins.get_registered_adapter(DefaultModule) is None:
    adapter_mixins.register_adapter(DefaultModule, DefaultModuleAdapter)


class TestAdapterMixin:
    @pytest.mark.unit
    def test_module_registered_adapter_by_class_path(self):
        classpath = get_classpath(DefaultModule)
        adapter_meta = adapter_mixins.get_registered_adapter(classpath)
        assert adapter_meta is not None
        assert adapter_meta.base_class == DefaultModule
        assert adapter_meta.adapter_class == DefaultModuleAdapter

    @pytest.mark.unit
    def test_module_registered_adapter_by_class(self):
        adapter_meta = adapter_mixins.get_registered_adapter(DefaultModule)
        assert adapter_meta is not None
        assert adapter_meta.base_class == DefaultModule
        assert adapter_meta.adapter_class == DefaultModuleAdapter

    @pytest.mark.unit
    def test_module_registered_adapter_by_adapter_class(self):
        adapter_meta = adapter_mixins.get_registered_adapter(DefaultModuleAdapter)
        assert adapter_meta is not None
        assert adapter_meta.base_class == DefaultModule
        assert adapter_meta.adapter_class == DefaultModuleAdapter

    @pytest.mark.unit
    def test_single_adapter(self):
        model = DefaultModuleAdapter()
        original_num_params = model.num_params()

        model.add_adapter(name='adapter_0', cfg=get_adapter_cfg())
        new_num_params = model.num_params()
        assert new_num_params > original_num_params

    @pytest.mark.unit
    def test_multiple_adapter(self):
        model = DefaultModuleAdapter()
        original_num_params = model.num_params()

        model.add_adapter(name='adapter_0', cfg=get_adapter_cfg())
        new_num_params = model.num_params()
        assert new_num_params > original_num_params

        original_num_params = new_num_params
        model.add_adapter(name='adapter_1', cfg=get_adapter_cfg())
        new_num_params = model.num_params()
        assert new_num_params > original_num_params

    @pytest.mark.unit
    def test_forward_linear_pre(self):
        torch.random.manual_seed(0)
        x = torch.randn(2, 50)

        model = DefaultModuleAdapter()
        origial_output = model(x)

        model.add_adapter(name='adapter_0', cfg=get_adapter_cfg())
        new_output = model(x)

        assert torch.mean(torch.abs(origial_output - new_output)) < 1e-5

    @pytest.mark.unit
    def test_forward_linear_post(self):
        torch.random.manual_seed(0)
        x = torch.randn(2, 50)

        model = DefaultModuleAdapter()
        origial_output = model(x)

        model.add_adapter(name='adapter_0', cfg=get_adapter_cfg(norm_pos='post'))
        new_output = model(x)

        assert torch.mean(torch.abs(origial_output - new_output)) < 1e-5

    @pytest.mark.unit
    def test_multi_adapter_forward(self):
        torch.random.manual_seed(0)
        x = torch.randn(2, 50)

        model = DefaultModuleAdapter()
        origial_output = model(x)

        model.add_adapter(name='adapter_0', cfg=get_adapter_cfg())
        model.add_adapter(name='adapter_1', cfg=get_adapter_cfg())
        new_output = model(x)

        assert model._adapter_names == ['adapter_0', 'adapter_1']
        assert torch.mean(torch.abs(origial_output - new_output)) < 1e-5

    @pytest.mark.unit
    def test_multi_adapter_partial_forward(self):
        torch.random.manual_seed(0)
        x = torch.randn(2, 50)

        model = DefaultModuleAdapter()
        origial_output = model(x)

        model.add_adapter(name='adapter_0', cfg=get_adapter_cfg())
        model.add_adapter(name='adapter_1', cfg=get_adapter_cfg())

        model.set_enabled_adapters(name='adapter_0', enabled=False)
        new_output = model(x)

        assert model._adapter_names == ['adapter_1']
        assert torch.mean(torch.abs(origial_output - new_output)) < 1e-5

    @pytest.mark.unit
    def test_forward_unfrozen_adapters(self):
        model = DefaultModuleAdapter()
        original_num_params = model.num_params()

        dim = 10
        model.add_adapter(name='adapter_0', cfg=get_adapter_cfg(dim=dim))
        model.freeze()
        model.unfreeze_enabled_adapters()

        assert original_num_params == 2650

        original_params = 0
        adapter_params = 0
        for name, param in model.named_parameters():
            if 'adapter' not in name:
                assert param.requires_grad is False
                original_params += param.numel()
            else:
                assert param.requires_grad is True
                adapter_params += param.numel()

        for mname, module in model.named_modules():
            if isinstance(module, (torch.nn.BatchNorm1d, torch.nn.BatchNorm2d, torch.nn.BatchNorm3d)):
                assert module.track_running_stats is False

        assert original_params > adapter_params

    @pytest.mark.unit
    def test_forward_linear_no_strategy(self):
        torch.random.manual_seed(0)
        x = torch.randn(2, 50)

        model = DefaultModuleAdapter()
        model.add_adapter(name='adapter_0', cfg=get_adapter_cfg())

        # delete the strategy
        adapter_module = model.adapter_layer[model.get_enabled_adapters()[0]]
        del adapter_module.adapter_strategy

        with pytest.raises(AttributeError):
            _ = model(x)

    @pytest.mark.unit
    def test_forward_linear_replaced_strategy(self):
        class MultiplyAdapterStrategy(adapter_mixin_strategies.AbstractAdapterStrategy):
            def forward(self, input: torch.Tensor, adapter: torch.nn.Module, *, module: AdapterModuleMixin):
                out = adapter(input)
                return input * out

        torch.random.manual_seed(0)
        x = torch.randn(2, 50)

        model = DefaultModuleAdapter()
        model.add_adapter(name='adapter_0', cfg=get_adapter_cfg())

        # modify the strategy
        adapter_module = model.adapter_layer[model.get_enabled_adapters()[0]]
        adapter_module.adapter_strategy = MultiplyAdapterStrategy()

        out = model(x)
        # result of adapter is zero tensor, output multiplied by adapter result should be zero
        assert (out > 0.0).any() == torch.tensor(False)
